import subprocess
import os
import sys


if snakemake.config["long_reads"] != "none":
    if snakemake.config["long_read_type"][0] in ["ont", "ont_hq"]:
        subprocess.Popen("coverm genome -t %d -d bins/final_bins/ -m relative_abundance covered_fraction --single %s -p minimap2-ont --min-covered-fraction 0.0 -x fna %s > data/long_abundances.tsv" %
                         (snakemake.threads, " ".join(snakemake.config["long_reads"]),
                          '--bam-file-cache-directory data/reads_mapped_to_mags/long/ --discard-unmapped' if snakemake.config['strain_analysis'] is True else ''), shell=True).wait()
    elif snakemake.config["long_read_type"][0] in ["rs", "sq", "ccs"]:
        subprocess.Popen("coverm genome -t %d -d bins/final_bins/ -m relative_abundance covered_fraction --single %s -p minimap2-pb --min-covered-fraction 0.0 -x fna %s > data/long_abundances.tsv" %
                         (snakemake.threads, " ".join(snakemake.config["long_reads"]),
                          '--bam-file-cache-directory data/reads_mapped_to_mags/long/ --discard-unmapped' if snakemake.config['strain_analysis'] is True else ''), shell=True).wait()
    else:
        subprocess.Popen("coverm genome -t %d -d bins/final_bins/ -m relative_abundance covered_fraction --single %s -p minimap2-ont --min-covered-fraction 0.0 -x fna %s > data/long_abundances.tsv" %
                         (snakemake.threads, " ".join(snakemake.config["long_reads"]),
                          '--bam-file-cache-directory data/reads_mapped_to_mags/long/ --discard-unmapped' if snakemake.config['strain_analysis'] is True else ''), shell=True).wait()

if snakemake.config['short_reads_2'] != 'none':
    subprocess.Popen("coverm genome -t %d -d bins/final_bins/ -m relative_abundance covered_fraction -1 %s -2 %s --min-covered-fraction 0.0 -x fna %s > data/short_abundances.tsv" %
                     (snakemake.threads, " ".join(snakemake.config["short_reads_1"]), " ".join(snakemake.config["short_reads_2"]),
                      '--bam-file-cache-directory data/reads_mapped_to_mags/short/ --discard-unmapped' if snakemake.config['strain_analysis'] is True else ''), shell=True).wait()

elif snakemake.config['short_reads_1'] != 'none':
    subprocess.Popen("coverm genome -t %d -d bins/final_bins/ -m relative_abundance covered_fraction --interleaved %s --min-covered-fraction 0.0 -x fna %s > data/short_abundances.tsv" %
                     (snakemake.threads, " ".join(snakemake.config["short_reads_1"]),
                      '--bam-file-cache-directory data/reads_mapped_to_mags/short/ --discard-unmapped' if snakemake.config['strain_analysis'] is True else ''), shell=True).wait()

# Concatenate the two coverage files if both long and short exist
if snakemake.config["long_reads"] != "none" and (snakemake.config["short_reads_1"] != "none"):
    with open('data/coverm_abundances.tsv', 'w') as file3:
        with open('data/short_abundances.tsv', 'r') as file1:
            with open('data/long_abundances.tsv', 'r') as file2:
                for line1, line2 in zip(file1, file2):
                    long_cov_line = "\t".join([l.strip() for l in line2.strip().split('\t')[1::]])
                    print(line1.strip(), "\t", long_cov_line, file=file3)
elif snakemake.config["long_reads"] != "none":  # rename long reads cov if only it exists
    os.rename("data/long_abundances.tsv", "data/coverm_abundances.tsv")
elif snakemake.config["short_reads_1"] != "none":  # rename shrot reads cov if only they exist
    os.rename("data/short_abundances.tsv", "data/coverm_abundances.tsv")



import sys
import os

os.makedirs('www/fraction_recovered', exist_ok=True)

if snakemake.config["long_reads"] != "none" and not os.path.exists("data/long_cov.tsv"):
    if snakemake.config["long_read_type"][0] in ["ont", "ont_hq"]:
        subprocess.Popen("coverm genome -t %d -r %s --single %s -p minimap2-ont --min-read-percent-identity 0.85 > www/fraction_recovered/long_fraction_recovered.tsv" %
                         (snakemake.threads, snakemake.input.fasta, " ".join(snakemake.config["long_reads"])), shell=True).wait()
    elif snakemake.config["long_read_type"][0] in ["rs", "sq", "ccs"]:
        subprocess.Popen("coverm genome -t %d -r %s --single %s -p minimap2-pb --min-read-percent-identity 0.85 > www/fraction_recovered/long_fraction_recovered.tsv" %
                         (snakemake.threads, snakemake.input.fasta, " ".join(snakemake.config["long_reads"])), shell=True).wait()
    else:
        subprocess.Popen(
            "coverm genome -t %d -r %s --single %s -p minimap2-ont > www/fraction_recovered/long_fraction_recovered.tsv" %
            (snakemake.threads, snakemake.input.fasta, " ".join(snakemake.config["long_reads"])), shell=True).wait()

if snakemake.config['short_reads_2'] != 'none' and not os.path.exists("data/short_cov.tsv"):
    subprocess.Popen("coverm genome -t %d -r %s -1 %s -2 %s > www/fraction_recovered/short_fraction_recovered.tsv" %
                     (snakemake.threads, snakemake.input.fasta, " ".join(snakemake.config["short_reads_1"]), " ".join(snakemake.config["short_reads_2"])), shell=True).wait()

elif snakemake.config['short_reads_1']  != 'none' and not os.path.exists("data/short_cov.tsv"):
    subprocess.Popen("coverm genome -t %d -r %s --interleaved %s > www/fraction_recovered/short_fraction_recovered.tsv" %
                     (snakemake.threads, snakemake.input.fasta, " ".join(snakemake.config["short_reads_1"])), shell=True).wait()


import pysam

samfile = pysam.AlignmentFile(snakemake.input[0], 'rb')


cutoff = 0.05
mapped_full = set()
overlap = set()
partial_map = set()
unmapped = set()
for read in samfile.fetch(until_eof=True):
    start = True
    if not read.cigartuples is None:
        clipped_start = 0
        clipped_end = 0
        for i in read.cigartuples:
            if i[0] == 4 or i[0] == 5:
                if start:
                    clipped_start += i[1]
                else:
                    clipped_end += i[1]
            else:
                start = False
        length = read.infer_read_length()
        if clipped_start/length <= cutoff and clipped_end/length <= cutoff:
            mapped_full.add(read.query_name)
        elif clipped_start/length <= cutoff and read.reference_end > read.reference_length - 100:
            overlap.add(read.query_name)
        elif clipped_end/length <= cutoff and read.reference_start < 100:
            overlap.add(read.query_name)
        else:
            partial_map.add(read.query_name)
    else:
        unmapped.add(read.query_name)


if snakemake.params[0] == 'no_full':
    all_reads = unmapped.union(partial_map).union(overlap)
    out_set = all_reads - mapped_full
with open(snakemake.output[0], 'w') as o:
    for i in out_set:
        o.write(i + '\n')


import subprocess
import os

if snakemake.config["long_reads"] != "none" and not os.path.exists("data/long_cov.tsv"):
    if snakemake.config["long_read_type"][0] in ["ont", "ont_hq"]:
        subprocess.Popen("TMPDIR=%s coverm contig -t %d -r %s --single %s -p minimap2-ont -m length trimmed_mean variance --bam-file-cache-directory data/binning_bams/ --discard-unmapped --min-read-percent-identity 0.85 > data/long_cov.tsv" %
                         (snakemake.params.tmpdir, snakemake.threads, snakemake.input.fasta, " ".join(snakemake.config["long_reads"])), shell=True).wait()
    elif snakemake.config["long_read_type"][0] in ["rs", "sq", "ccs"]:
        subprocess.Popen("TMPDIR=%s coverm contig -t %d -r %s --single %s -p minimap2-pb -m length trimmed_mean variance --bam-file-cache-directory data/binning_bams/ --discard-unmapped --min-read-percent-identity 0.9 > data/long_cov.tsv" %
                         (snakemake.params.tmpdir, snakemake.threads, snakemake.input.fasta, " ".join(snakemake.config["long_reads"])), shell=True).wait()
    else:
        subprocess.Popen(
            "TMPDIR=%s coverm contig -t %d -r %s --single %s -p minimap2-ont -m length trimmed_mean variance --bam-file-cache-directory data/binning_bams/ --discard-unmapped > data/long_cov.tsv" %
            (snakemake.params.tmpdir, snakemake.threads, snakemake.input.fasta, " ".join(snakemake.config["long_reads"])), shell=True).wait()

if snakemake.config['short_reads_2'] != 'none' and not os.path.exists("data/short_cov.tsv"):
    subprocess.Popen("TMPDIR=%s coverm contig -t %d -r %s -1 %s -2 %s -m metabat --bam-file-cache-directory data/binning_bams/ --discard-unmapped > data/short_cov.tsv" %
                     (snakemake.params.tmpdir, snakemake.threads, snakemake.input.fasta, " ".join(snakemake.config["short_reads_1"]), " ".join(snakemake.config["short_reads_2"])), shell=True).wait()

elif snakemake.config['short_reads_1']  != 'none' and not os.path.exists("data/short_cov.tsv"):
    subprocess.Popen("TMPDIR=%s coverm contig -t %d -r %s --interleaved %s -m metabat --bam-file-cache-directory data/binning_bams/ --discard-unmapped > data/short_cov.tsv" %
                     (snakemake.params.tmpdir, snakemake.threads, snakemake.input.fasta, " ".join(snakemake.config["short_reads_1"])), shell=True).wait()

# subprocess.Popen("ls data/binning_bams/*.bam | parallel -j1 samtools index -@ %d {} {}.bai" % (snakemake.threads-1), shell=True).wait()
# Concatenate the two coverage files if both long and short exist
if snakemake.config["long_reads"] != "none" and (snakemake.config["short_reads_1"] != "none"):
    short_count = len(snakemake.config["short_reads_1"])
    long_count = len(snakemake.config["long_reads"])
    with open('data/coverm.cov', 'w') as file3:
        with open('data/short_cov.tsv', 'r') as file1:
            with open('data/long_cov.tsv', 'r') as file2:
                for idx, (line1, line2) in enumerate(zip(file1, file2)):
                    long_values = line2.strip().split("\t")[1::]
                    del long_values[0::3] # delete length value
                    short_values = line1.strip().split("\t")
                    if idx != 0:
                        long_cov = sum([float(x) for x in long_values[0::2]])
                        short_cov = sum([float(x) for x in short_values[3::2]])
                        tot_depth = (long_cov + short_cov) / (short_count + long_count)
                        line = (
                            "{contig_name}\t"
                            "{contig_length}\t"
                            "{total_avg_depth}\t"
                            "{short}\t"
                            "{long}"
                        ).format(
                            contig_name = short_values[0],
                            contig_length = short_values[1],
                            total_avg_depth = tot_depth,
                            short = "\t".join(short_values[3::]),
                            long = "\t".join(long_values)
                        )
                        print(line, file=file3)
                    else:
                        line = (
                            "{short}\t{long}"
                        ).format(
                            short = "\t".join(short_values),
                            long = "\t".join(long_values)
                        )
                        print(line, file=file3)
                        
elif snakemake.config["long_reads"] != "none":
    with open('data/coverm.cov', 'w') as file3:
        with open('data/long_cov.tsv', 'r') as file1:
            for idx, line1 in enumerate(file1):
                long_values = line1.strip().split("\t")
                del long_values[4::3] # delete extra length values
                long_count = len(long_values[2::2])
                if idx != 0:
                    long_cov = sum([float(x) for x in long_values[2::2]])
                    tot_depth = long_cov / long_count
                    line = (
                        "{contig_name}\t"
                        "{contig_length}\t"
                        "{total_avg_depth}\t"
                        "{long}"
                    ).format(
                        contig_name = long_values[0],
                        contig_length = long_values[1],
                        total_avg_depth = tot_depth,
                        long = "\t".join(long_values[2::])
                    )
                    print(line, file=file3)
                else:
                    line = (
                        "contigName\tcontigLen\ttotalAvgDepth\t{samples}"
                    ).format(
                        samples = "\t".join(long_values[2::])
                    )
                    print(line, file=file3)
                        
elif snakemake.config["short_reads_1"] != "none":  # rename shrot reads cov if only they exist
    os.rename("data/short_cov.tsv", "data/coverm.cov")

if snakemake.config["long_reads"] != "none":
    with open('data/long.cov', 'w') as file3:
        with open('data/long_cov.tsv', 'r') as file:
            for idx, line1 in enumerate(file):
                long_values = line1.strip().split("\t")
                del long_values[4::3] # delete extra length values
                long_count = len(long_values[2::2])
                if idx != 0:
                    long_cov = sum([float(x) for x in long_values[2::2]])
                    tot_depth = long_cov / long_count
                    line = (
                        "{contig_name}\t"
                        "{contig_length}\t"
                        "{total_avg_depth}\t"
                        "{long}"
                    ).format(
                        contig_name = long_values[0],
                        contig_length = long_values[1],
                        total_avg_depth = tot_depth,
                        long = "\t".join(long_values[2::])
                    )
                    print(line, file=file3)
                else:
                    line = (
                        "contigName\tcontigLen\ttotalAvgDepth\t{samples}"
                    ).format(
                        samples = "\t".join(long_values[2::])
                    )
                    print(line, file=file3)
        # os.remove("data/long_cov.tsv")

try:
    os.makedirs("data/maxbin_cov/")
except OSError:
    pass
with open("data/coverm.cov") as f, open("data/maxbin.cov.list", "w") as o:
    cov_list = []
    contig_list = []
    f.readline()
    for line in f:
        contig = line.split()[0]
        contig_list.append(contig)
        depths = line.split()[3:]
        cov_list.append([])
        for i in range(0, len(depths), 2):
            cov_list[-1].append(depths[i])
    for i in range(len(cov_list[0])):
        with open("data/maxbin_cov/p%d.cov" % i, 'w') as oo:
            for j in range(len(contig_list)):
                oo.write(contig_list[j] + '\t' + cov_list[j][i] + '\n')
        o.write("data/maxbin_cov/p%d.cov\n" % i)
# with open("data/binning_bams/done", 'w') as o:
#     o.write('done')







#!/usr/bin/env python
###############################################################################
# processor.py - Class used to generate config file and make calls to the
#                snakemake pipeline
###############################################################################
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the                #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program. If not, see <http://www.gnu.org/licenses/>.        #
#                                                                             #
###############################################################################
import glob

import aviary.config.config as Config
__author__ = "Rhys Newell"
__copyright__ = "Copyright 2020"
__credits__ = ["Rhys Newell"]
__license__ = "GPL3"
__maintainer__ = "Rhys Newell"
__email__ = "rhys.newell near hdr.qut.edu.au"
__status__ = "Development"

###############################################################################
# System imports
import sys
import logging
import os
import subprocess
from pathlib import Path
from glob import glob

# Local imports
from snakemake import utils
from snakemake.io import load_configfile
from ruamel.yaml import YAML  # used for yaml reading with comments

# Debug
debug={1:logging.CRITICAL,
       2:logging.ERROR,
       3:logging.WARNING,
       4:logging.INFO,
       5:logging.DEBUG}

###############################################################################
############################### - Exceptions - ################################

class BadTreeFileException(Exception):
    pass

###############################################################################
################################ - Functions - ################################

def get_snakefile(file="Snakefile"):
    sf = os.path.join(os.path.dirname(os.path.abspath(__file__)), file)
    if not os.path.exists(sf):
        sys.exit("Unable to locate the Snakemake workflow file; tried %s" % sf)
    return sf


# def update_config(config):
#     """
#     Populates config file with default config values.
#     And made changes if necessary.
#     """
#
#     # get default values and update them with values specified in config file
#     default_config = make_default_config()
#     utils.update_config(default_config, config)
#
#     return default_config


###############################################################################
################################ - Classes - ##################################

class Processor:
    def __init__(self,
                 args,
                 # conda_prefix=Config.get_software_db_path('CONDA_ENV_PATH', '--conda-prefix'),
                 ):


        self.conda_prefix = args.conda_prefix
        self.tmpdir = os.path.abspath(args.tmpdir)
        self.resources = args.resources
        self.output = os.path.abspath(args.output)
        self.threads = args.max_threads
        self.max_memory = args.max_memory
        self.pplacer_threads = min(int(self.threads), 48)
        self.workflows = args.workflow

        try:
            self.strain_analysis = args.strain_analysis
        except AttributeError:
            self.strain_analysis = False

        # binning group items
        try:
            self.min_contig_size = args.min_contig_size
            self.min_bin_size = args.min_bin_size
            self.semibin_model = args.semibin_model

            if args.skip_binners is not None:
                # skipping binners is tricky and requires that bam indices are produced
                # so we explicitly place the bam indices rule in the DAG before any binning occurs
                self.skip_binners = [binner.lower() for binner in args.skip_binners]
                self.check_binners_to_skip()
                if not any(rule in self.workflows for rule in ["get_bam_indices", "complete_assembly", "complete_assembly_with_qc", "complete_cluster"]):
                    self.workflows.insert(0, 'get_bam_indices')
                elif any(rule in self.workflows for rule in ["complete_assembly", "complete_assembly_with_qc"]):
                    indices = [idx for idx, rule in enumerate(self.workflows) if rule in ["complete_assembly", "complete_assembly_with_qc"]]
                    if len(indices) > 1:
                        logging.critical(f"Rules 'complete_assembly' and 'complete_assembly_with_qc' both found in DAG or one found multiple times.")
                        logging.critical(f"Please revise your --workflow parameter.")
                        sys.exit(1)
                    self.workflows.insert(indices[0] + 1, 'get_bam_indices')

        except AttributeError:
            self.min_contig_size = 1500
            self.min_bin_size = 200000
            self.semibin_model = 'global'
            self.skip_binners = []

        try:
            self.assembly = args.assembly
        except AttributeError:
            self.assembly = 'none'

        try:
            self.reference_filter = os.path.abspath(args.reference_filter) if args.reference_filter != 'none' else 'none'
            if args.gold_standard is not None:
                self.gold_standard = [os.path.abspath(p) for p in args.gold_standard]
            else:
                self.gold_standard = 'none'
        except AttributeError:
            self.reference_filter = 'none'
            self.gold_standard = 'none'

        try:
            self.gsa_mappings = args.gsa_mappings
        except AttributeError:
            self.gsa_mappings = 'none'

        try:
            self.longreads = args.longreads
            self.long_percent_identity = args.long_percent_identity
        except AttributeError:
            self.longreads = 'none'
            self.long_percent_identity = 'none'

        try:
            self.longread_type = args.longread_type
        except AttributeError:
            self.longread_type = 'none'

        try:
            self.short_percent_identity = args.short_percent_identity

            if args.coupled != "none":
                self.pe1 = args.coupled[::2]
                self.pe2 = args.coupled[1::2]
                if len(self.pe1) != len(self.pe2):
                    logging.error(f"Number of forward reads != Number of reverse reads. Current forward: {len(self.pe1)} reverse: {len(self.pe2)}")
                    sys.exit(-1)
            else:
                self.pe2 = args.pe2
                if args.interleaved == "none":
                    self.pe1 = args.pe1
                elif args.pe2 == "none" and args.interleaved != "none":
                    self.pe1 = args.interleaved
        except AttributeError:
            self.pe1 = 'none'
            self.pe2 = 'none'
            self.short_percent_identity = 'none'


        try:
            self.kmer_sizes = args.kmer_sizes
            self.use_megahit = args.use_megahit
            self.coassemble = args.coassemble
        except AttributeError:
            self.kmer_sizes = ['auto']
            self.use_megahit = False
            self.coassemble = True

        try:
            self.mag_directory = os.path.abspath(args.directory) if args.directory is not None else 'none'
        except AttributeError:
            self.mag_directory = 'none'

        try:
            if args.gtdb_path is not None:
                self.gtdbtk = args.gtdb_path
            else:
                self.gtdbtk = Config.get_software_db_path('GTDBTK_DATA_PATH', '--gtdb-path')
            if args.eggnog_db_path is not None:
                self.eggnog = args.eggnog_db_path
            else:
                self.eggnog = Config.get_software_db_path('EGGNOG_DATA_DIR', '--eggnog-db-path')
        except AttributeError:
            self.gtdbtk = Config.get_software_db_path('GTDBTK_DATA_PATH', '--gtdb-path')
            self.eggnog = Config.get_software_db_path('EGGNOG_DATA_DIR', '--eggnog-db-path')
            # self.enrichm = Config.get_software_db_path('ENRICHM_DB', '--enrichm-db-path')

        try:
            self.mag_extension = args.ext
        except AttributeError:
            self.mag_extension = 'none'

        try:
            self.previous_runs = [os.path.abspath(run) for run in args.previous_runs]
        except AttributeError:
            self.previous_runs = 'none'

        # Aviary cluster arguments
        try:
            if args.min_completeness == 'none':
                self.min_completeness = ' '
            else:
                self.min_completeness = f'--min-completeness {args.min_completeness}'

            if args.max_contamination == 'none':
                self.max_contamination = ' '
            else:
                self.max_contamination = f'--max-contamination {args.max_contamination}'

            self.precluster_ani = fraction_to_percent(args.precluster_ani)
            self.ani = fraction_to_percent(args.ani)
            self.precluster_method = args.precluster_method
            self.use_checkm2_scores = args.use_checkm2_scores
            self.pggb_params = args.pggb_params
        except AttributeError:
            self.min_completeness = 'none'
            self.max_contamination = 'none'
            self.ani = 'none'
            self.precluster_ani = 'none'
            self.precluster_method = 'none'
            self.use_checkm2_scores = False
            self.pggb_params = 'none'

        try:
            if args.checkm2_db_path is not None:
                self.checkm2_db = args.checkm2_db_path
            else:
                self.checkm2_db = Config.get_software_db_path('CHECKM2DB', '--checkm2-db-path')
        except AttributeError:
            self.checkm2_db = Config.get_software_db_path('CHECKM2DB', '--checkm2-db-path')
            # self.checkm2_db = 'none'

        # Must be always be first workflow
        if args.download:
            self.workflows.insert(0, 'download_databases')


    def make_config(self):
        """
        Reads template config file with comments from ./template_config.yaml
        updates it by the parameters provided.
        """

        self.config = os.path.join(self.output, 'config.yaml')

        yaml = YAML()
        yaml.version = (1, 1)
        yaml.default_flow_style = False

        template_conf_file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                          "template_config.yaml")

        with open(template_conf_file) as template_config:
            conf = yaml.load(template_config)

        if self.assembly != "none" and self.assembly is not None:
            self.assembly = list(set([os.path.abspath(p) for p in self.assembly]))
        elif self.assembly is None:
            self.assembly = 'none'
            logging.warning("No assembly provided, assembly will be created using available reads...")
        if self.pe1 != "none":
            self.pe1 = list(set([os.path.abspath(p) for p in self.pe1]))
        if self.pe2 != "none":
            self.pe2 = list(set([os.path.abspath(p) for p in self.pe2]))
        if self.longreads != "none":
            self.longreads = list(([os.path.abspath(p) for p in self.longreads]))
        if self.gsa_mappings != "none":
            self.gsa_mappings = os.path.abspath(self.gsa_mappings)

        conf["fasta"] = self.assembly
        conf["reference_filter"] = self.reference_filter
        conf["gsa"] = self.gold_standard
        conf["gsa_mappings"] = self.gsa_mappings
        conf["semibin_model"] = self.semibin_model
        conf["max_threads"] = int(self.threads)
        conf["pplacer_threads"] = int(self.pplacer_threads)
        conf["max_memory"] = int(self.max_memory)
        conf["short_reads_1"] = self.pe1
        conf["short_reads_2"] = self.pe2
        conf["long_reads"] = self.longreads
        conf["long_read_type"] = self.longread_type
        conf["kmer_sizes"] = self.kmer_sizes
        conf["use_megahit"] = self.use_megahit
        conf["coassemble"] = self.coassemble
        conf["min_contig_size"] = int(self.min_contig_size)
        conf["min_bin_size"] = int(self.min_bin_size)
        conf["gtdbtk_folder"] = self.gtdbtk
        conf["eggnog_folder"] = self.eggnog
        conf["strain_analysis"] = self.strain_analysis
        conf["checkm2_db_folder"] = self.checkm2_db
        conf["use_checkm2_scores"] = self.use_checkm2_scores
        conf["mag_directory"] = self.mag_directory
        conf["mag_extension"] = self.mag_extension
        conf["previous_runs"] = self.previous_runs
        conf["min_completeness"] = self.min_completeness
        conf["max_contamination"] = self.max_contamination
        conf["ani"] = self.ani
        conf["precluster_ani"] = self.precluster_ani
        conf["precluster_method"] = self.precluster_method
        conf["pggb_params"] = self.pggb_params
        conf["tmpdir"] = self.tmpdir

        with open(self.config, "w") as f:
            yaml.dump(conf, f)
        logging.info(
            "Configuration file written to %s\n" % self.config
        )

    def _validate_config(self):
        load_configfile(self.config)

    def check_binners_to_skip(self):

        to_skip = []
        skipped = 0

        if "rosella" in self.skip_binners:
            rosella_path = self.output + "/data/rosella_bins/"
            rosella_ref_path = self.output + "/data/rosella_refined/"
            os.makedirs(rosella_path, exist_ok=True)
            os.makedirs(rosella_ref_path, exist_ok=True)
            to_skip.append(rosella_path + "done")
            to_skip.append(rosella_ref_path + "done")
            skipped += 1

        if "semibin" in self.skip_binners:
            sb_path = self.output + "/data/semibin_bins/"
            sb_ref_path = self.output + "/data/semibin_refined/"
            os.makedirs(sb_path, exist_ok=True)
            os.makedirs(sb_ref_path, exist_ok=True)
            to_skip.append(sb_path + "done")
            to_skip.append(sb_ref_path + "done")
            skipped += 1

        if "vamb" in self.skip_binners:
            vamb_path = self.output + "/data/vamb_bins/"
            os.makedirs(vamb_path, exist_ok=True)
            to_skip.append(vamb_path + "done")
            skipped += 1

        if "concoct" in self.skip_binners:
            concoct_path = self.output + "/data/concoct_bins/"
            os.makedirs(concoct_path, exist_ok=True)
            to_skip.append(concoct_path + "done")
            skipped += 1

        if any(m in self.skip_binners for m in ["maxbin", "maxbin2"]):
            maxbin2_path = self.output + "/data/maxbin2_bins/"
            os.makedirs(maxbin2_path, exist_ok=True)
            to_skip.append(maxbin2_path + "done")
            skipped += 1

        if any(m in self.skip_binners for m in ["metabat", "metabat1"]):
            to_skip.extend(self.skip_metabat1())
            skipped += 1

        if any(m in self.skip_binners for m in ["metabat", "metabat2"]):
            m_path = self.output + "/data/metabat_bins_2/"
            m_ref_path = self.output + "/data/metabat2_refined/"
            os.makedirs(m_path, exist_ok=True)
            os.makedirs(m_ref_path, exist_ok=True)
            to_skip.append(m_path + "done")
            to_skip.append(m_ref_path + "done")
            skipped += 1

        if skipped < 7:
            [Path(skip).touch(exist_ok=True) for skip in to_skip]
        else:
            logging.error("Check --skip-binners. All binners are being skipped. At least one binning algorithm must be used.")
            sys.exit(1)


    def skip_metabat1(self):
        return_paths = []
        for m in ["metabat_bins_sens/", "metabat_bins_ssens/", "metabat_bins_spec/", "metabat_bins_sspec/"]:
            m_path = self.output + "/data/" + m
            os.makedirs(m_path, exist_ok=True)
            return_paths.append(m_path + "done")

        return return_paths

    def run_workflow(self, cores=16, profile=None,
                     dryrun=False, clean=True, conda_frontend="mamba",
                     snakemake_args="", write_to_script=None):
        """
        Runs the aviary pipeline
        By default all steps are executed
        Needs a config-file which is generated by given inputs.
        Most snakemake arguments can be appended to the command for more info see 'snakemake --help'
        """

        if not os.path.exists(self.config):
            logging.critical(f"config-file not found: {self.config}\n")
            sys.exit(1)

        self._validate_config()

        cores = max(int(self.threads), cores)

        for workflow in self.workflows:
            cmd = (
                "snakemake --snakefile {snakefile} --directory {working_dir} "
                "{jobs} --rerun-incomplete {args} "
                "--configfile '{config_file}' --nolock "
                "{profile} {conda_frontend} {resources} --use-conda {conda_prefix} "
                "{dryrun} {notemp} "
                "{target_rule}"
            ).format(
                snakefile=get_snakefile(),
                working_dir=self.output,
                jobs="--jobs {}".format(cores) if cores is not None else "",
                config_file=self.config,
                profile="" if (profile is None) else "--profile {}".format(profile),
                dryrun="--dryrun" if dryrun else "",
                notemp="--notemp" if not clean else "",
                args=snakemake_args,
                target_rule=workflow if workflow != "None" else "",
                conda_prefix="--conda-prefix " + self.conda_prefix,
                conda_frontend="--conda-frontend " + conda_frontend,
                resources=f"--default-resources \"tmpdir='{self.tmpdir}'\" --resources mem_mb={int(self.max_memory)*1024} {self.resources}" if not dryrun else ""
            )

            if write_to_script is not None:
                write_to_script.append(cmd)
                continue

            try:
                subprocess.check_call(cmd, shell=True)
                logging.info("Executing: %s" % cmd)
            except subprocess.CalledProcessError as e:
                # removes the traceback
                logging.critical(e)
                exit(1)


def process_batch(args, prefix):
    '''
    Function for handling and processing aviary commands in batches
    The user supplies a tab separated batch file with six defined columns. Aviary is run on each line
    of the batch file using the specified workflow.

    If the user wishes, the results are then clustered.
    '''
    import pandas as pd

    logging.info(f"Reading batch file: {args.batch_file}")

    header=0
    with open(args.batch_file, mode='r') as check_batch:
        for line in check_batch.readlines():
            if "sample\tshort_reads_1\tshort_reads_2\tlong_reads\tlong_read_type\tassembly\tcoassemble" in line \
                or "sample,short_reads_1,short_reads_2,long_reads,long_read_type,assembly,coassemble" in line \
                or "sample  short_reads_1   short_reads_2   long_reads      long_read_type  assembly        coassemble" in line \
                or "sample short_reads_1 short_reads_2 long_reads long_read_type assembly coassemble" in line:
               header=1
               logging.debug("Inferred header")
            else:
                logging.debug("No heading inferred.")
            break

    batch = pd.read_csv(args.batch_file, sep=None, engine='python', skiprows=header)
    if len(batch.columns) != 7:
        logging.critical(f"Batch file contains incorrect number of columns ({len(batch.columns)}). Should contain 7.")
        logging.critical(f"Current columns: {batch.columns}")
        sys.exit()

    if args.build:
        try:
            args.cmds = args.cmds + '--conda-create-envs-only '
        except TypeError:
            args.cmds = '--conda-create-envs-only '

    if args.use_unicycler:
        args.workflow.insert(0, "combine_assemblies")

    try:
        script_file = args.write_script
        write_to_script = []
    except AttributeError:
        script_file = None
        write_to_script = None

    runs = []
    args.interleaved = "none" # hacky solution to skip attribute error
    args.coupled = "none"
    for i in range(batch.shape[0]):
        # process the batch line
        sample = check_batch_input(batch.iloc[i, 0], f"sample_{i}", split=False)
        logging.info(f"Processing {sample}")
        s1 = check_batch_input(batch.iloc[i, 1], "none", split=True)
        s2 = check_batch_input(batch.iloc[i, 2], "none", split=True)
        l = check_batch_input(batch.iloc[i, 3], "none", split=True)
        l_type = check_batch_input(batch.iloc[i, 4], "ont", split=False)
        assembly = check_batch_input(batch.iloc[i, 5], None, split=False)
        coassemble = check_batch_input(batch.iloc[i, 6], False, split=False)

        # update the value of args
        args.output = f"{prefix}/{sample}"
        runs.append(args.output)
        args.pe1 = s1
        args.pe2 = s2

        args.longreads = l
        args.longread_type = l_type
        args.assembly = assembly
        args.coassemble = coassemble

        # ensure output folder exists
        if not os.path.exists(args.output):
            os.makedirs(args.output)

        # setup processor for this line
        processor = Processor(args)
        processor.make_config()

        processor.run_workflow(cores=int(args.n_cores),
                               dryrun=args.dryrun,
                               clean=args.clean,
                               conda_frontend=args.conda_frontend,
                               snakemake_args=args.cmds,
                               write_to_script=write_to_script)

    if args.cluster:
        logging.info(f"Beginning clustering of {len(runs)} previous Aviary runs with ANI values: {args.ani_values}...")

        for ani in args.ani_values:
            args.previous_runs = runs
            args.ani = ani
            args.workflow = ['complete_cluster']
            args.output = f"{prefix}/aviary_cluster_ani_{ani}"
            # ensure output folder exists
            if not os.path.exists(args.output):
                os.makedirs(args.output)
            processor = Processor(args)
            processor.make_config()

            processor.run_workflow(cores=int(args.n_cores),
                                   dryrun=args.dryrun,
                                   clean=args.clean,
                                   conda_frontend=args.conda_frontend,
                                   snakemake_args=args.cmds,
                                   write_to_script=write_to_script)

    if script_file is not None:
        with open(script_file, 'w') as sf:
            for line in write_to_script:
                sf.write(f"{line}\n")


def check_batch_input(val, default=None, split=False, split_val=','):
    """
    Takes a batch entry from within a line and ensures the output makes sense for Aviary
    Split is required for cells that separated by the split_val
    """
    if not isinstance(val, str):
        return default

    new_val = val.strip()

    if split:
        new_val = new_val.split(split_val)
        return_vals = []
        for paths in new_val:
            return_vals.extend(glob(paths))
        new_val = return_vals

    return new_val

def fraction_to_percent(val):
    if val <= 1:
        return val * 100
    return val


import subprocess
import os
import logging


if not os.path.exists(snakemake.output.fasta):
    try:
        subprocess.Popen("cat %s %s > %s" % (snakemake.input.flye_fasta,
                                             snakemake.input.short_fasta,
                                             snakemake.output.fasta), shell=True).wait()
        logging.info("Flye and metaSPAdes/Unicycler assemblies combined...")
    except AttributeError:
        subprocess.Popen("ln %s %s" % (snakemake.input.fasta,
                                       snakemake.output.fasta), shell=True).wait()
        logging.info("Treating Flye assembly as final assembly...")


import subprocess
import os

long_reads = None
short_reads = None

if os.path.isdir("data/reads_mapped_to_mags/long/"):
    long_reads = "-l data/reads_mapped_to_mags/long/*.bam"
elif snakemake.config["long_reads"] != "none":
    long_reads = f"--longreads {' '.join(snakemake.config['long_reads'])}"
else:
    long_reads = ""

if os.path.isdir("data/reads_mapped_to_mags/short/"):
    short_reads = "-b data/reads_mapped_to_mags/short/*.bam"
elif snakemake.config["short_reads_2"] != "none":
    short_reads = f"-1 {' '.join(snakemake.config['short_reads_1'])} -2 {' '.join(snakemake.config['short_reads_2'])}"
elif snakemake.config["short_reads_1"] != "none":
    short_reads = f"--interleaved {' '.join(snakemake.config['short_reads_1'])}"
else:
    short_reads = ""

subprocess.Popen(f"lorikeet call -t {snakemake.threads} -d {snakemake.config['mag_directory']} "
                 f"-x {snakemake.params.mag_extension} -p {snakemake.params.parallel_genomes} {short_reads} {long_reads} "
                 f"-o {snakemake.output.output_directory} --calculate-dnds --calculate-fst", shell=True).wait()



import os
import subprocess

if snakemake.params.long_read_type == 'ont':
    subprocess.Popen(
        "flye --nano-raw %s --meta -o data/flye -t %d" %
        (snakemake.input.fastq, snakemake.threads), shell=True).wait()
elif snakemake.params.long_read_type == 'ont_hq':
    subprocess.Popen(
        "flye --nano-hq %s --meta -o data/flye -t %d" %
        (snakemake.input.fastq, snakemake.threads), shell=True).wait()
elif snakemake.params.long_read_type == 'ccs':
    subprocess.Popen(
        "flye --pacbio-hifi %s --meta -o data/flye -t %d" %
        (snakemake.input.fastq, snakemake.threads), shell=True).wait()
else:
    subprocess.Popen(
        "flye --pacbio-raw %s --meta -o data/flye -t %d" %
        (snakemake.input.fastq, snakemake.threads), shell=True).wait()


import os
import signal
import subprocess

"""
Function to handle signal IOErrors after missing input
"""
def handler(signum, frame):
     raise IOError
signal.signal(signal.SIGALRM, handler)


def source_conda_env():
    try:
        with open(format('%s/etc/conda/activate.d/aviary.sh' % os.environ['CONDA_PREFIX'])) as f:
            for line in f:
                if line.startswith('#') or not line.strip():
                    continue
                # if 'export' not in line:
                #     continue
                # Remove leading `export `, if you have those
                # then, split name / value pair
                # key, value = line.replace('export ', '', 1).strip().split('=', 1)
                try:
                    key, value = line.strip().split('=', 1)
                    key = key.strip('export ')
                    # print(key, value)
                    os.environ[key] = value  # Load to local environ
                except ValueError:
                    continue
    except FileNotFoundError:
        # File not found so going to have to create it
        pass

def source_bashrc():
    try:
        with open('%s/.bashrc' % os.environ['HOME']) as f:
            for line in f:
                if line.startswith('#') or not line.strip():
                    continue
                # if 'export' not in line:
                #     continue
                # Remove leading `export `, if you have those
                # then, split name / value pair
                # key, value = line.replace('export ', '', 1).strip().split('=', 1)
                try:
                    key, value = line.strip().split('=', 1)
                    key = key.strip('export ')
                    os.environ[key] = value  # Load to local environ
                except ValueError:
                    continue
    except FileNotFoundError:
        # File not found so going to have to create it
        pass

"""
Load the reference package. This will fail if the directory doesn't exist.
"""
def get_software_db_path(db_name='CONDA_ENV_PATH', software_flag='--conda-prefix'):
    try:
        source_conda_env()
        SW_PATH = os.environ[db_name]
        return SW_PATH
    except KeyError:
        try:
            source_conda_env()
            CONDA_PATH = os.environ[db_name]
            return CONDA_PATH
        except KeyError:
            try:
                source_bashrc()
                CONDA_PATH = os.environ[db_name]
                return CONDA_PATH
            except KeyError:
                print('\n' + '=' * 100)
                print(' ERROR '.center(100))
                print('_' * 100 + '\n')
                print(f"The '{db_name}' environment variable is not defined.".center(100) + '\n')
                print(f'Please set this variable to your default server/home directory containing {db_name}.'.center(100))
                print(f'Alternatively, use {software_flag} flag.'.center(100))
                print(f'Note: This variable must point to the DIRECTORY containing the files, not the files themselves'.center(100))
                print('=' * 100)
                signal.alarm(120)
                os.environ[db_name] = input(f'Input path to directory for {db_name} now:').strip()
                try:
                    subprocess.Popen(
                        'mkdir -p %s/etc/conda/activate.d/; mkdir -p %s/etc/conda/deactivate.d/; echo "export %s=%s" >> %s/etc/conda/activate.d/aviary.sh; echo "unset %s" >> %s/etc/conda/deactivate.d/aviary.sh; ' %
                        (os.environ['CONDA_PREFIX'], os.environ['CONDA_PREFIX'], db_name, os.environ[db_name],
                         os.environ['CONDA_PREFIX'], db_name, os.environ['CONDA_PREFIX']), shell=True).wait()
                except KeyError:
                    subprocess.Popen(
                        'echo "export %s=%s" >> ~/.bashrc' %
                        (db_name, os.environ[db_name]), shell=True).wait()
                signal.alarm(0)
                print('=' * 100)
                # print('Reactivate your aviary conda environment or source ~/.bashrc to suppress this message.'.center(100))
                # print('=' * 100)

                return os.environ[db_name]


"""
Sets an environmental variable and appends it to the conda activation script
"""
def set_db_path(path, db_name='CONDA_ENV_PATH'):
    os.environ[db_name] = path.strip()
    try:
        subprocess.Popen(
            'mkdir -p %s/etc/conda/activate.d/; mkdir -p %s/etc/conda/deactivate.d/; echo "export %s=%s" >> %s/etc/conda/activate.d/aviary.sh; echo "unset %s" >> %s/etc/conda/deactivate.d/aviary.sh; ' %
            (os.environ['CONDA_PREFIX'], os.environ['CONDA_PREFIX'], db_name, os.environ[db_name],
             os.environ['CONDA_PREFIX'], db_name, os.environ['CONDA_PREFIX']), shell=True).wait()
    except KeyError:
        subprocess.Popen(
            'echo "export %s=%s" >> ~/.bashrc' %
            (db_name, os.environ[db_name]), shell=True).wait()



#!/usr/bin/env python
###############################################################################
# 
###############################################################################
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the                #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program. If not, see <http://www.gnu.org/licenses/>.        #
#                                                                             #
###############################################################################
__author__ = "Rhys Newell"
__copyright__ = "Copyright 2020"
__credits__ = ["Rhys Newell"]
__version__ = "1.0.0"
__license__ = "GPL3"
__maintainer__ = "Rhys Newell"
__email__ = "rhys.newell near hdr.qut.edu.au"
__status__ = "Development"

###############################################################################
# System imports
import sys
import argparse
import logging
import os
import datetime
from pathlib import Path

# Function imports
import numpy as np
import pandas as pd
from Bio import SeqIO

# Debug
debug = {
    1: logging.CRITICAL,
    2: logging.ERROR,
    3: logging.WARNING,
    4: logging.INFO,
    5: logging.DEBUG
}

###############################################################################
############################### - Exceptions - ################################


class BadTreeFileException(Exception):
    pass


###############################################################################                                                                                                                      [44/1010]
################################ - Functions - ################################
def main():
    ############################ ~ Main Parser ~ ##############################
    vamb_options = argparse.ArgumentParser(
        'write_vamb_bins',
        description='Bin out the results of vamb',
        formatter_class=CustomHelpFormatter,
        epilog='''
                                        ~ write_vamb_bins ~
            How to use write_vamb_bins:

            write_vamb_bins --reference assembly.fasta --clusters vamb_clusters.tsv

            NOTE: Provided defaults are for use with aviary and should be changed for your specific use

            ''')
    vamb_options.add_argument('--version',
                             action='version',
                             version=__version__,
                             help='Show version information.')
    vamb_options.add_argument(
        '--verbosity',
        help=
        '1 = critical, 2 = error, 3 = warning, 4 = info, 5 = debug. Default = 4 (logging)',
        type=int,
        default=4)
    vamb_options.add_argument('--log',
                             help='Output logging information to file',
                             default=False)

    ########################## ~ sub-parser ~ #########################

    vamb_options.add_argument('--reference',
                              help='The assembly file to be binned',
                              dest='assembly',
                              default="data/vamb_bams/renamed_contigs.fasta")

    vamb_options.add_argument('--clusters',
                              help='The vamb clusters',
                              dest='clusters',
                              default="data/vamb_bins/clusters.tsv")

    vamb_options.add_argument('--min_size',
                              help='Minimum bin size',
                              dest='min_size',
                              default=200000)

    vamb_options.add_argument('--output',
                              help='The output directory',
                              dest='output',
                              default='data/vamb_bins/')

    vamb_options.set_defaults(func=vamb)
    ###########################################################################
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Parsing input ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

    args = vamb_options.parse_args()
    time = datetime.datetime.now().strftime('%H:%M:%S %d-%m-%Y')

    if args.log:
        if os.path.isfile(args.log):
            raise Exception("File %s exists" % args.log)
        logging.basicConfig(
            filename=args.log,
            level=debug[args.verbosity],
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%m/%d/%Y %I:%M:%S %p')
    else:
        logging.basicConfig(
            level=debug[args.verbosity],
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%m/%d/%Y %I:%M:%S %p')

    logging.info("Time - %s" % (time))
    logging.info("Command - %s" % ' '.join(sys.argv))

    args.func(args)


def vamb(args):
    min_bin_size = int(args.min_size)
    prefix = args.output
    if not os.path.exists(prefix):
        os.makedirs(prefix)

    bins = {}
    with open(args.clusters, 'r') as vamb_file:
        for line in vamb_file:
            line = line.split()
            try:
                bins[line[0]].append(line[1])
            except KeyError:
                bins[line[0]] = [line[1]]

    assembly = SeqIO.to_dict(SeqIO.parse(args.assembly, "fasta"))

    logging.info("Writing bins...")
    max_cluster_id = max(bins.keys())
    for (bin, contigs) in bins.items():
        if bin != -1:
            # Calculate total bin size and check if it is larger than min_bin_size
            bin_length = sum([len(assembly[contig].seq) for contig in contigs])
            if bin_length >= min_bin_size:
                with open(prefix + '/vamb_bin.' + str(bin) + '.fna', 'w') as f:
                    for contig in contigs:
                        write_contig(contig, assembly, f)

        else:
            # Get final bin value
            max_cluster_id += 1
            # Rescue any large unbinned contigs and put them in their own cluster
            for contig in contigs:
                if len(assembly[contig].seq) >= min_bin_size:
                    with open(prefix + '/vamb_bin.' + str(max_cluster_id) + '.fna', 'w') as f:
                        write_contig(contig, assembly, f)

    Path('data/vamb_bins/done').touch()


def write_contig(contig, assembly, f):
    seq = assembly[contig]
    fasta = ">" + seq.id + '\n'
    fasta += str(seq.seq) + '\n'
    f.write(fasta)

###############################################################################
################################ - Classes - ##################################


class CustomHelpFormatter(argparse.HelpFormatter):
    def _split_lines(self, text, width):
        return text.splitlines()

    def _get_help_string(self, action):
        h = action.help
        if '%(default)' not in action.help:
            if action.default != '' and \
               action.default != [] and \
               action.default != None \
               and action.default != False:
                if action.default is not argparse.SUPPRESS:
                    defaulting_nargs = [
                        argparse.OPTIONAL, argparse.ZERO_OR_MORE
                    ]

                    if action.option_strings or action.nargs in defaulting_nargs:

                        if '\n' in h:
                            lines = h.splitlines()
                            lines[0] += ' (default: %(default)s)'
                            h = '\n'.join(lines)
                        else:
                            h += ' (default: %(default)s)'
        return h

    def _fill_text(self, text, width, indent):
        return ''.join([indent + line for line in text.splitlines(True)])



if __name__ == '__main__':

   sys.exit(main())



import subprocess
import os
import sys

try:
    os.mkdir("data/singlem_out/")
except OSError:
    print("Using prexisting directory: data/singlem_out/")

if snakemake.config["long_reads"] != "none":
        subprocess.Popen(
            "singlem pipe --threads %d --sequences %s --otu_table data/singlem_out/metagenome.longread_otu_table.csv" %
            (snakemake.config["pplacer_threads"], " ".join(snakemake.config["long_reads"])), shell=True).wait()


if snakemake.config["short_reads_2"] != "none":
        subprocess.Popen(
            "singlem pipe --threads %d --forward %s --reverse %s --otu_table data/singlem_out/metagenome.shortread_otu_table.csv" %
            (snakemake.config["pplacer_threads"],
             " ".join(snakemake.config["short_reads_1"]),
             " ".join(snakemake.config["short_reads_2"])), shell=True).wait()
elif snakemake.config["short_reads_1"] != "none":
        subprocess.Popen(
            "singlem pipe --threads %d --sequences %s --otu_table data/singlem_out/metagenome.shortread_otu_table.csv" %
            (snakemake.config["pplacer_threads"],
             " ".join(snakemake.config["short_reads_1"])), shell=True).wait()

subprocess.Popen(
            "singlem summarise --input_otu_tables data/singlem_out/*.csv --output_otu_table data/singlem_out/metagenome.combined_otu_table.csv", shell=True).wait()


import subprocess
import sys

read_set1 = snakemake.config['short_reads_1']
read_set2 = snakemake.config['short_reads_2']

# deal with read sets i.e. are we coassembling? Which assembler are we using?
# Non co-assembled reads are handled the same for each assembler
if read_set1 != 'none':
    if not snakemake.params.coassemble or len(read_set1) == 1:
        read_set1 = read_set1[0]
        if read_set2 != 'none':
            read_set2 = read_set2[0]

    elif not snakemake.params.use_megahit:
        if read_set2 != 'none':
            for reads1, reads2 in zip(read_set1, read_set2):
                subprocess.Popen(f"cat {reads1} >> data/short_reads.1.fastq.gz", shell=True).wait()
                subprocess.Popen(f"cat {reads2} >> data/short_reads.2.fastq.gz", shell=True).wait()
            read_set1 = 'data/short_reads.1.fastq.gz'
            read_set2 = 'data/short_reads.2.fastq.gz'
        else:
            for reads1 in read_set1:
                subprocess.Popen(f"cat {reads1} >> data/short_reads.1.fastq.gz", shell=True).wait()
            read_set1 = 'data/short_reads.1.fastq.gz'

    else:
        read_set1 = ",".join(read_set1)
        if read_set2 != 'none':
            read_set2 = ",".join(read_set2)
else:
    # forward reads must always be present as either
    # forward or interleaved or single ended
    print(f"============= ERROR =============\n")
    print(f"Invalid read sets provided \n "
          f"for short read assembly: \n")
    print(f"    Set 1: {read_set1}\n")
    print(f"    Set 2: {read_set2}\n")
    print("Validate read sets and resubmit")
    sys.exit(1)


# designate input read string
read_string = f"--12 {read_set1}"
if read_set2 != 'none':
    read_string = f"-1 {read_set1} -2 {read_set2}"


# Run chosen assembler
if snakemake.params.use_megahit:
    command = f"megahit {read_string} -t {snakemake.threads} -m {snakemake.config['max_memory']} " \
              f"-o data/megahit_assembly --tmp-dir {snakemake.params.tmpdir}"
    print(f"Queueing command {command}")

    subprocess.Popen(
        command,
        shell=True
    ).wait()
    subprocess.Popen(
        "mkdir -p data/short_read_assembly; cp data/megahit_assembly/final.contigs.fa data/short_read_assembly/scaffolds.fasta",
        shell=True
    ).wait()
else:
    kmers = " ".join(snakemake.params.kmer_sizes)
    command = f"spades.py --memory {snakemake.config['max_memory']} --meta -t {snakemake.threads} " \
              f"-o data/short_read_assembly {read_string} -k {kmers} --tmp-dir {snakemake.params.tmpdir}"
    print(f"Queueing command {command}")

    subprocess.Popen(
        command,
        shell=True).wait()


import subprocess
import os
import multiprocessing as mp
import glob
import logging

def process_short_reads_interleaved_filtered(reads, file):
    subprocess.Popen('seqtk seq -1 %s | mfqe --fastq-read-name-lists %s --output-fastq-files %s.1.fastq.gz' %
                     (reads, file, file[:-5]), shell=True).wait()
    subprocess.Popen('seqtk seq -2 %s | mfqe --fastq-read-name-lists %s --output-fastq-files %s.2.fastq.gz' %
                     (reads, file, file[:-5]), shell=True).wait()

def process_short_reads_interleaved(reads, file, threads):
    subprocess.Popen('seqtk seq -1 %s | seqkit -j %d grep -f %s | pigz -p %d > %s.1.fastq.gz' %
                     (reads, threads, file, threads, file[:-5]), shell=True).wait()
    subprocess.Popen('seqtk seq -2 %s | seqkit -j %d grep -f %s | pigz -p %d > %s.2.fastq.gz' %
                     (reads, threads, file, threads, file[:-5]), shell=True).wait()

def process_short_reads_paired(reads_1, reads_2, file, threads):
    subprocess.Popen('seqkit -j %d grep -f %s %s | pigz -p %d > %s.1.fastq.gz' %
                     (threads, file, reads_1, threads, file[:-5]), shell=True).wait()
    subprocess.Popen('seqkit -j %d grep -f %s %s | pigz -p %d > %s.2.fastq.gz' %
                     (threads, file, reads_2, threads, file[:-5]), shell=True).wait()

def process_long_reads(reads, file, threads):
    subprocess.Popen('seqtk subseq %s %s | pigz -p %d > %s.fastq.gz' %
                     (reads, file, threads, file[:-5]), shell=True).wait()


def get_index(n_files, current):
    if current >= n_files / 2:
        return 2
    else:
        return 1


if snakemake.config['long_reads'] != 'none':
    logging.info("Extracting binned long reads...")

    pool = mp.Pool(snakemake.threads)

    mp_results = [pool.apply_async(process_long_reads,
                                    args=(snakemake.config["long_reads"][0], file, 2))
                                    for file in glob.glob('data/binned_reads/*.long.list')]
    for result in mp_results:
        result.get()

    pool.close()
    pool.join()

if os.path.exists('data/short_reads.fastq.gz'):
    logging.info("Extracting binned interleaved filtered short reads...")
    pool = mp.Pool(snakemake.threads)
    # extract reads from filtered short reads
    # take end 1 and 2 separately
    mp_results = [pool.apply_async(process_short_reads_interleaved_filtered,
                                    args=('data/short_reads.fastq.gz', file))
                                    for file in glob.glob('data/binned_reads/*.short.list')]
    for result in mp_results:
        result.get()

    pool.close()
    pool.join()

elif snakemake.config['short_reads_2'] != 'none':
    logging.info("Extracting binned paired end short reads...")
    pool = mp.Pool(snakemake.threads // 2)
    # extract reads from all short read files
    mp_results = [pool.apply_async(process_short_reads_paired,
                                    args=(' '.join(snakemake.config['short_reads_1']),
                                          ' '.join(snakemake.config['short_reads_2']),
                                          file, 2))
                                    for file in glob.glob('data/binned_reads/*.short.list')]

    for result in mp_results:
        result.get()

    pool.close()
    pool.join()

elif snakemake.config['short_reads_1'] != 'none':
    logging.info("Extracting binned interleaved short reads...")
    pool = mp.Pool(snakemake.threads // 2)
    # extract reads from all short read files
    mp_results = [pool.apply_async(process_short_reads_interleaved,
                                   args=(' '.join(snakemake.config['short_reads_1']),
                                         file, 2))
                  for file in glob.glob('data/binned_reads/*.short.list')]

    for result in mp_results:
        result.get()

    pool.close()
    pool.join()

subprocess.Popen('touch %s' % (snakemake.output), shell=True).wait()




import os
import subprocess


try:
    os.makedirs("data/final_assemblies")
except FileExistsError:
    pass


out_assemblies = []
with open(snakemake.input.list) as f:
    for line in f:
        if len(line.split()) == 6:
            mb_bin, long_reads, length, bases_nano, short_reads, bases_ill = line.split()
        else:
            mb_bin, long_reads, length, bases_nano = line.split()
            short_reads, bases_ill = 'none', 0
        if os.stat(long_reads).st_size == 0:
            no_long = True
        else:
            no_long = False
        long_reads = long_reads[:-5] + '.fastq.gz'
        long_reads = os.path.abspath(long_reads)
        short_reads_1 = short_reads[:-5] + '.1.fastq.gz'
        short_reads_2 = short_reads[:-5] + '.2.fastq.gz'
        short_reads_1 = os.path.abspath(short_reads_1)
        short_reads_2 = os.path.abspath(short_reads_2)
        length, bases_nano, bases_ill = float(length), float(bases_nano), float(bases_ill)
        out_assemblies.append('data/final_assemblies/%s_unicyc/assembly.fasta' % mb_bin)
        if not os.path.exists('data/final_assemblies/%s_unicyc/assembly.fasta' % mb_bin):
            if short_reads == 'none':
                subprocess.Popen("unicycler --verbosity 0 -t %d -l %s -o data/final_assemblies/%s_unicyc" % (
                    snakemake.threads, long_reads, mb_bin), shell=True).wait()
            elif no_long:
                subprocess.Popen("unicycler --verbosity 0 -t %d -1 %s -2 %s -o data/final_assemblies/%s_unicyc" % (
                snakemake.threads, short_reads_1, short_reads_2, mb_bin), shell=True).wait()
            else:
                subprocess.Popen("unicycler --verbosity 0 -t %d -1 %s -2 %s -l %s -o data/final_assemblies/%s_unicyc" % (
                    snakemake.threads, short_reads_1, short_reads_2, long_reads, mb_bin), shell=True).wait()

unbinned_set = set()
if os.path.exists(snakemake.input.metabat_done[:-4] + "binned_contigs.unbinned"):
    with open(snakemake.input.metabat_done[:-4] + "binned_contigs.unbinned") as f:
        for line in f:
            unbinned_set.add(line.rstrip())


with open(snakemake.output.fasta, 'w') as o:
    count = 0
    getseq = False
    with open(snakemake.input.fasta) as f:
        for line in f:
            if line.startswith('>') and line.split()[0][1:] in unbinned_set:
                getseq = True
                o.write('>unbinned_' + str(count) + '\n')
                count += 1
            elif line.startswith('>'):
                getseq = False
            elif getseq:
                o.write(line)
    for i in out_assemblies:
        if not os.path.exists(i):
            with open(i[:-14] + 'unicycler.log') as f:
                lastline = f.readlines()[-1]
                if lastline.startswith("Error: SPAdes failed to produce assemblies. See spades_assembly/assembly/spades.log for more info.") or \
                    lastline.startswith("Error: none of the SPAdes graphs were suitable for scaffolding in Unicycler") or \
                    lastline.startswith("Error: miniasm assembly failed"):
                    continue
        with open(i) as assembly:
            for line in assembly:
                if line.startswith('>'):
                    count += 1
                    o.write('>unicycler_' + str(count) + '\n')
                else:
                    o.write(line)

if not os.path.exists(snakemake.output.fasta):
    open(snakemake.output.fasta, 'a').close()


import subprocess
import os
import sys


def process_batch(batch_file_path):
    main_directory = os.getcwd()
    install_directory = "~/git/aviary"
    with open(batch_file_path, "r") as batch_file:
        for line in batch_file:
            line = line.strip().split()
            assembly = line[0]
            identifier = line[1]
            # Assuming that batch file has reads written as interleaved files e.g. 1 2 1 2...
            # File paths must also be absolute
            forward = line[2::2]
            reverse = line[3::2]
            try:
                os.mkdir("data/" + identifier)
            except FileExistsError:
                print("Directory already exists for sample %s" % identifier)
            # Symbolically link to main aviary folder === BAD IDEA ===
            # subprocess.Popen("ln -s %s/annotation.smk %s/data/%s/"
            #                  % (main_directory, main_directory, identifier), shell=True).wait()
            # subprocess.Popen("cp %s/template_config.yaml %s/data/%s/"
                             # % (main_directory, main_directory, identifier), shell=True).wait()
            subprocess.Popen("ln -s %s/envs %s/data/%s/"
                             % (main_directory, main_directory, identifier), shell=True).wait()
            subprocess.Popen("ln -s %s/scripts %s/data/%s/"
                             % (main_directory, main_directory, identifier), shell=True).wait()
            subprocess.Popen("ln -s %s/.snakemake/ %s/data/%s/"
                             % (main_directory, main_directory, identifier), shell=True).wait()

            # Flags to specify when to change the next line
            changing_fasta = False
            changing_reads_1 = False
            changing_reads_2 = False
            changing_batch = False
            with open("%s/data/%s/template_config.yaml" % (main_directory, identifier), "w+") as new_config_file:
                with open("%s/template_config.yaml" % main_directory, 'r') as config_file:
                    for config_line in config_file:
                        if config_line.startswith("fasta"):
                            new_config_file.write(config_line)
                            changing_fasta = True
                        elif changing_fasta:
                            new_config_file.write(" " + assembly + "\n")
                            changing_fasta = False
                        elif config_line.startswith("short_reads_1"):
                            new_config_file.write(config_line)
                            changing_reads_1 = True
                        elif changing_reads_1:
                            config_line = " " + " ".join(forward) + "\n"
                            new_config_file.write(config_line)
                            changing_reads_1 = False
                        elif config_line.startswith("short_reads_2"):
                            new_config_file.write(config_line)
                            changing_reads_2 = True
                        elif changing_reads_2:
                            config_line = " " + " ".join(reverse) + "\n"
                            new_config_file.write(config_line)
                            changing_reads_2 = False
                        elif config_line.startswith("batch_file"):
                            new_config_file.write(config_line)
                            changing_batch = True
                        elif changing_batch:
                            config_line = " " + "none" + "\n"
                            new_config_file.write(config_line)
                            changing_batch = False
                        else:
                            new_config_file.write(config_line)


            os.chdir("%s/data/%s" % (main_directory, identifier))
            # Run a new snakemake process using the updated template_config.yaml
            subprocess.Popen("snakemake --unlock --use-conda --conda-prefix %s/.snakemake/ -s %s/annotation.smk recover_mags"
                             % (install_directory, install_directory), shell=True).wait()
            subprocess.Popen("snakemake --use-conda --conda-prefix %s/.snakemake/ -s %s/annotation.smk --cores %d recover_mags"
                             % (install_directory, install_directory, snakemake.threads), shell=True).wait()
            os.chdir(main_directory)


if __name__ == "__main__":
    process_batch(snakemake.config["batch_file"])
    subprocess.Popen("touch data/done", shell=True).wait()


import subprocess
from pathlib import Path
import os

if os.path.exists('data/short_reads.fastq.gz'):
    subprocess.Popen("minimap2 -ax sr -t %d %s data/short_reads.fastq.gz | samtools view -b -f 12 -@ %d > %s; "
                     "samtools index -@ %d %s; "
                     "samtools bam2fq -@ %d -f 12 %s | pigz -p %d > %s" %
                     (snakemake.threads, snakemake.input.reference_filter,
                      snakemake.threads, snakemake.output.bam,
                      snakemake.threads, snakemake.output.bam,
                      snakemake.threads, snakemake.output.bam,
                      snakemake.threads, snakemake.output.fastq,
                      ), shell=True).wait()

elif snakemake.config['short_reads_2'] != 'none':
    if len(snakemake.config['short_reads_2']) == 1 or not snakemake.params.coassemble:
        pe1 = snakemake.config['short_reads_1'][0]
        pe2 = snakemake.config['short_reads_2'][0]
    else:
        if not os.path.exists("data/short_reads.1.fastq.gz"):
            for reads1, reads2 in zip(snakemake.config['short_reads_1'], snakemake.config['short_reads_2']):
                subprocess.Popen(f"cat {reads1} >> data/short_reads.1.fastq.gz", shell=True).wait()
                subprocess.Popen(f"cat {reads2} >> data/short_reads.2.fastq.gz", shell=True).wait()
        pe1 = "data/short_reads.1.fastq.gz"
        pe2 = "data/short_reads.2.fastq.gz"

    subprocess.Popen("minimap2 -ax sr -t %d %s %s %s | samtools view -b -f 12 -@ %d > %s; "
                     "samtools index -@ %d %s; "
                     "samtools bam2fq -@ %d -f 12 %s | pigz -p %d > %s" %
                     (snakemake.threads, snakemake.input.reference_filter, pe1, pe2,
                      snakemake.threads, snakemake.output.bam,
                      snakemake.threads, snakemake.output.bam,
                      snakemake.threads, snakemake.output.bam,
                      snakemake.threads, snakemake.output.fastq,
                      ), shell=True).wait()

elif snakemake.config['short_reads_1']  != 'none':
    if len(snakemake.config['short_reads_1']) == 1 or not snakemake.params.coassemble:
        pe1 = snakemake.config['short_reads_1'][0]
    else:
        if not os.path.exists("data/short_reads.1.fastq.gz"):
            for reads1 in snakemake.config['short_reads_1']:
                subprocess.Popen(f"cat {reads1} >> data/short_reads.1.fastq.gz", shell=True).wait()
        pe1 = "data/short_reads.1.fastq.gz"

    subprocess.Popen("minimap2 -ax sr -t %d %s %s | samtools view -b -f 12 -@ %d > %s; "
                     "samtools index -@ %d %s; "
                     "samtools bam2fq -@ %d -f 12 %s | pigz -p %d > %s" %
                     (snakemake.threads, snakemake.input.reference_filter, pe1,
                      snakemake.threads, snakemake.output.bam,
                      snakemake.threads, snakemake.output.bam,
                      snakemake.threads, snakemake.output.bam,
                      snakemake.threads, snakemake.output.fastq,
                      ), shell=True).wait()

Path(snakemake.output.filtered).touch()


import subprocess
import multiprocessing as mp
import os
import logging


def spawn_fastqc(file, threads=1):
    subprocess.Popen('fastqc -o www/fastqc/ -t %d %s' % (threads, file), shell=True).wait()


if os.path.exists('data/short_reads.fastq.gz'):
    subprocess.Popen("fastqc -o www/fastqc/ -t %d data/short_reads.fastq.gz" % (snakemake.threads))
elif snakemake.config['short_reads_2'] != 'none': # paired end
    pool = mp.Pool(snakemake.threads)
    if isinstance(snakemake.config['short_reads_1'], str):
        threads = snakemake.threads
        reads = [snakemake.config['short_reads_1'], snakemake.config['short_reads_2']]
    else:
        threads = max(len(snakemake.config['short_reads_1'] + snakemake.config['short_reads_2']) // snakemake.threads, 1)
        reads = snakemake.config['short_reads_1'] + snakemake.config['short_reads_2']
    mp_results = [pool.apply_async(spawn_fastqc, args=(read, threads))
                  for read in
                  reads]

    for result in mp_results:
        result.get()

    pool.close()
    pool.join()
elif snakemake.config['short_reads_1'] != 'none': # interleaved
    pool = mp.Pool(snakemake.threads)
    if isinstance(snakemake.config['short_reads_1'], str):
        threads = snakemake.threads
        reads = [snakemake.config['short_reads_1']]
    else:
        threads = max(len(snakemake.config['short_reads_1']) // snakemake.threads,
                      1)
        reads = snakemake.config['short_reads_1']

    mp_results = [pool.apply_async(spawn_fastqc, args=(read, threads))
                  for read in
                  reads]

    for result in mp_results:
        result.get()

    pool.close()
    pool.join()
else:
    subprocess.Popen('echo "no short reads" > www/fastqc/short_reads_fastqc.html', shell=True).wait()

subprocess.Popen('touch www/fastqc/done', shell=True).wait()



#!/usr/bin/env python
###############################################################################
# aviary.py - Info about aviary.py
###############################################################################
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the                #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program. If not, see <http://www.gnu.org/licenses/>.        #
#                                                                             #
###############################################################################
import aviary.config.config as Config
from aviary.modules.processor import Processor, process_batch
from .__init__ import __version__
__author__ = "Rhys Newell"
__copyright__ = "Copyright 2022"
__credits__ = ["Rhys Newell"]
__license__ = "GPL3"
__maintainer__ = "Rhys Newell"
__email__ = "rhys.newell near hdr.qut.edu.au"
__status__ = "Development"

###############################################################################
# System imports
import sys
import argparse
import logging
import os
from datetime import datetime
import tempfile

# Debug
debug={1:logging.CRITICAL,
       2:logging.ERROR,
       3:logging.WARNING,
       4:logging.INFO,
       5:logging.DEBUG}

###############################################################################
############################### - Exceptions - ################################

class BadTreeFileException(Exception):
    pass

###############################################################################
################################ - Functions - ################################
def centerify(text, width=-1):
  lines = text.split('\n')
  width = max(map(len, lines)) if width == -1 else width
  return '\n'.join(line.center(width) for line in lines)


def phelp():
    print(
"""

                    ......:::::: AVIARY ::::::......

           A comprehensive metagenomics bioinformatics pipeline

Metagenome assembly, binning, and annotation:
        assemble  - Perform hybrid assembly using short and long reads, 
                    or assembly using only short reads
        recover   - Recover MAGs from provided assembly using a variety 
                    of binning algorithms 
        annotate  - Annotate MAGs using EggNOG and GTBD-tk
        diversity - Perform strain diversity analysis of MAGs using Lorikeet
        complete  - Runs each stage of the pipeline: assemble, recover, 
                    annotate, diversity in that order.
        cluster   - Combines and dereplicates the MAGs from multiple Aviary runs
                    using Galah
        batch     - Run Aviary using a given workflow on a supplied batch of samples
                    and cluster the end result.

Isolate assembly, binning, and annotation:
        isolate   - Perform isolate assembly **PARTIALLY COMPLETED**
        
Utility modules:
        configure - Set or overwrite the environment variables for future runs.

"""
)


def str2bool(v):
    if isinstance(v, bool):
        return(v)
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return(True)
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return(False)
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

def main():
    if len(sys.argv) == 1 or sys.argv[1] == '-h' or sys.argv[1] == '--help':
        phelp()
        return

    # Source the conda environment variables in case users have previously set
    # the variables using config but have not restarted the environment.
    try:
        Config.source_conda_env()
    except FileNotFoundError:
        Config.source_bashrc()

    ############################ ~ Main Parser ~ ##############################
    main_parser = argparse.ArgumentParser(prog='aviary',
                                          formatter_class=CustomHelpFormatter,
                                          add_help=False)
    main_parser.add_argument('--version',
                             action='version',
                             version=__version__,
                             help='Show version information.')
    main_parser.add_argument('--verbosity',
                             help='1 = critical, 2 = error, 3 = warning, 4 = info, 5 = debug. Default = 4 (logging)',
                             type=int,
                             default=4)
    main_parser.add_argument('--log',
                             help='Output logging information to file',
                             default=False)
    subparsers = main_parser.add_subparsers(help="--", dest='subparser_name')

    #~#~#~#~#~#~#~#~#~#~#~#~#~ Command groups ~#~#~#~#~#~#~#~#~#~#~#~#~#

    ####################################################################

    base_group = argparse.ArgumentParser(formatter_class=CustomHelpFormatter,
                                         add_help=False)

    base_group.add_argument(
        '-t', '--max-threads', '--max_threads',
        help='Maximum number of threads given to any particular process',
        dest='max_threads',
        default=8,
    )

    base_group.add_argument(
        '-p', '--pplacer-threads', '--pplacer_threads',
        help=argparse.SUPPRESS,
        dest='pplacer_threads',
        default=8,
    )

    base_group.add_argument(
        '-n', '--n-cores', '--n_cores',
        help='Maximum number of cores available for use. Must be >= to max_threads',
        dest='n_cores',
        default=16,
    )

    base_group.add_argument(
        '-m', '--max-memory', '--max_memory',
        help='Maximum memory for available usage in Gigabytes',
        dest='max_memory',
        default=250,
    )

    base_group.add_argument(
        '-o', '--output',
        help='Output directory',
        dest='output',
        default='./',
    )

    base_group.add_argument(
        '--conda-prefix', '--conda_prefix',
        help='Path to the location of installed conda environments, or where to install new environments. \n'
             'Can be configured within the `configure` subcommand',
        dest='conda_prefix',
        default=None,
    )

    base_group.add_argument(
        '--tmpdir', '--tempdir', '--tmp-dir', '--tmp_dir', '--tmp', '--temp', '--temp-dir', '--temp_dir',
        help='Path to the location that will be treated used for temporary files. If none is specified, the TMPDIR \n'
             'environment variable will be used. Can be configured within the `configure` subcommand',
        dest='tmpdir',
        default=tempfile.gettempdir(),
    )

    base_group.add_argument(
        '--default-resources',
        help='Snakemake resources used as is found at: https://snakemake.readthedocs.io/en/stable/snakefiles/rules.html?highlight=resources#standard-resources \n'
             'NOTE: tmpdir is handled by the `tmpdir` command line parameter. ',
        dest='resources',
        default=" "
    )

    base_group.add_argument(
        '--dry-run', '--dry_run', '--dryrun',
        help='Perform snakemake dry run, tests workflow order and conda environments',
        type=str2bool,
        nargs='?',
        const=True,
        dest='dryrun',
        default=False,
    )

    base_group.add_argument(
        '--conda-frontend', '--conda_frontend',
        help='Which conda frontend to use, mamba is faster but harder to debug. Switch this to conda \n'
             'If experiencing problems installing environments',
        dest='conda_frontend',
        default="mamba",
        choices=["conda", "mamba"],
    )

    base_group.add_argument(
        '--clean',
        help='Clean up all temporary files. This will remove most BAM files and any FASTQ files \n'
             'generated from read filtering. Setting this to False is the equivalent of the --notemp \n'
             'option in snakemake. Useful for when running only part of a workflow as it avoids \n'
             'deleting files that would likely be needed in later parts of the workflow. \n'
             'NOTE: Not cleaning makes reruns faster but will incur the wrath of your sysadmin',
        type=str2bool,
        nargs='?',
        const=True,
        dest='clean',
        default=True,
    )

    base_group.add_argument(
        '--build',
        help='Build conda environments and then exits. Equivalent to \"--snakemake-cmds \'--conda-create-envs-only True \' \"',
        type=str2bool,
        nargs='?',
        const=True,
        dest='build',
    )

    base_group.add_argument(
        '--download', '--download',
        help='Downloads the required GTDB, EggNOG, & CheckM2 databases if required',
        type=str2bool,
        nargs='?',
        const=True,
        dest='download',
    )

    base_group.add_argument(
        '--snakemake-cmds',
        help='Additional commands to supplied to snakemake in the form of a single string'
             'e.g. "--print-compilation True". \n '
             'NOTE: Most commands in snakemake -h are valid but some commands may clash with commands \n '
             'aviary directly supplies to snakemake. Please make sure your additional commands don\'t clash.',
        dest='cmds',
        default='',
    )

    ####################################################################
    qc_group = argparse.ArgumentParser(formatter_class=CustomHelpFormatter,
                                               add_help=False)

    qc_group.add_argument(
        '-g', '--gold-standard-assembly', '--gold_standard_assembly',
        help='Gold standard assembly to compare either the Aviary assembly or a given input assembly against',
        dest="gold_standard",
        default=['none'],
        nargs='*'
    )

    qc_group.add_argument(
        '--gsa-mappings', '--gsa_mappings',
        help='CAMI I & II GSA mappings',
        dest="gsa_mappings",
        default='none'
    )

    qc_group.add_argument(
        '-r', '--reference-filter', '--reference_filter',
        help='Reference filter file to aid in the assembly',
        dest="reference_filter",
        nargs=1,
        default='none'
    )

    qc_group.add_argument(
        '--min-read-size', '--min_read_size',
        help='Minimum long read size when filtering using Filtlong',
        dest="min_read_size",
        default=250
    )

    qc_group.add_argument(
        '--min-mean-q', '--min_mean_q',
        help='Minimum mean quality threshold',
        dest="min_mean_q",
        default=50
    )

    qc_group.add_argument(
        '--keep-percent', '--keep_percent',
        help='Percentage of reads passing quality thresholds kept by filtlong',
        dest="keep_percent",
        default=100
    )


    ####################################################################

    short_read_group = argparse.ArgumentParser(formatter_class=CustomHelpFormatter,
                                         add_help=False)
    read_group_exclusive = short_read_group.add_mutually_exclusive_group()

    read_group_exclusive.add_argument(
        '-1', '--pe-1', '--paired-reads-1', '--paired_reads_1', '--pe1',
        help='A space separated list of forwards read files \n'
             'NOTE: If performing assembly and multiple files and longreads \n'
             '      are provided then only the first file will be used for assembly. \n'
             '      If no longreads are provided then all samples will be co-assembled \n'
             '      with megahit or metaspades depending on the --coassemble parameter\n',
        dest='pe1',
        nargs='*',
        default="none"
    )

    short_read_group.add_argument(
        '-2', '--pe-2', '--paired-reads-2', '--paired_reads_2', '--pe2',
        help='A space separated list of reverse read files \n'
             'NOTE: If performing assembly and multiple files and longreads \n'
             '      are provided then only the first file will be used for assembly. \n'
             '      If no longreads are provided then all samples will be co-assembled \n'
             '      with megahit or metaspades depending on the --coassemble parameter',
        dest='pe2',
        nargs='*',
        default="none"
    )

    read_group_exclusive.add_argument(
        '-i','--interleaved',
        help='A space separated list of interleaved read files \n'
             'NOTE: If performing assembly and multiple files and longreads \n'
             '      are provided then only the first file will be used for assembly. \n'
             '      If no longreads are provided then all samples will be co-assembled \n'
             '      with megahit or metaspades depending on the --coassemble parameter',
        dest='interleaved',
        nargs='*',
        default="none"
    )

    read_group_exclusive.add_argument(
        '-c', '--coupled',
        help='Forward and reverse read files in a coupled space separated list. \n'
             'NOTE: If performing assembly and multiple files and longreads \n'
             '      are provided then only the first file will be used for assembly. \n'
             '      If no longreads are provided then all samples will be co-assembled \n'
             '      with megahit or metaspades depending on the --coassemble parameter',
        dest='coupled',
        nargs='*',
        default="none"
    )

    read_group_exclusive.add_argument(
        '--min-percent-read-identity-short', '--min_percent_read_identity_short',
        help='Minimum percent read identity used by CoverM for short-reads \n'
             'when calculating genome abundances.',
        dest='short_percent_identity',
        default='95'
    )

    ####################################################################

    long_read_group = argparse.ArgumentParser(formatter_class=CustomHelpFormatter,
                                              add_help=False)
    long_read_group.add_argument(
        '-l', '--longreads', '--long-reads', '--long_reads',
        help='A space separated list of long-read read files. '
             'NOTE: If performing assembly and multiple long read files are provided, \n'
             '      then only the first file is used for assembly. This behaviour might change in future.',
        dest='longreads',
        nargs='*',
        default="none"
    )

    long_read_group.add_argument(
        '-z', '--longread-type', '--longread_type', '--long_read_type', '--long-read-type',
        help='Whether the sequencing platform and technology for the longreads. \n'
             '"rs" for PacBio RSII, "sq" for PacBio Sequel, "ccs" for PacBio CCS \n'
             'reads, "ont" for Oxford Nanopore and "ont_hq" for Oxford Nanopore high quality reads (Guppy5+ or Q20) \n',
        dest='longread_type',
        nargs=1,
        default="ont",
        choices=["ont","ont_hq", "rs", "sq", "ccs"],
    )

    long_read_group.add_argument(
        '--min-percent-read-identity-long', '--min_percent_read_identity_long',
        help='Minimum percent read identity used by CoverM for long-reads'
             'when calculating genome abundances.',
        dest='long_percent_identity',
        default='85'
    )

    ####################################################################

    annotation_group = argparse.ArgumentParser(add_help=False)

    # annotation_group.add_argument(
    #     '--enrichm-db-path', '--enrichm_db_path',
    #     help='Path to the local EnrichM Database files',
    #     dest='enrichm_db_path',
    #     default=Config.get_software_db_path('ENRICHM_DB', '--enrichm-db-path'),
    # )

    annotation_group.add_argument(
        '--gtdb-path', '--gtdb_path',
        help='Path to the local gtdb database files',
        dest='gtdb_path',
        default=None,
    )

    annotation_group.add_argument(
        '--eggnog-db-path', '--eggnog_db_path',
        help='Path to the local eggnog database files',
        dest='eggnog_db_path',
        default=None,
    )

    annotation_group.add_argument(
        '--checkm2-db-path', '--checkm2_db_path',
        help='Path to Checkm2 Database',
        dest='checkm2_db_path',
        required=False,
    )

    ####################################################################

    binning_group = argparse.ArgumentParser(formatter_class=CustomHelpFormatter,
                                            add_help=False)

    binning_group.add_argument(
        '-s', '--min-contig-size', '--min_contig_size',
        help='Minimum contig size in base pairs to be considered for binning',
        dest='min_contig_size',
        default=1500
    )

    binning_group.add_argument(
        '-b', '--min-bin-size', '--min_bin_size',
        help='Minimum bin size in base pairs for a MAG',
        dest='min_bin_size',
        default=200000
    )

    binning_group.add_argument(
        '--semibin-model', '--semibin_model',
        help='The environment model to passed to SemiBin. Can be one of: \n'
             'human_gut, dog_gut, ocean, soil, cat_gut, human_oral, mouse_gut, pig_gut, built_environment, wastewater, global',
        dest='semibin_model',
        default='global'
    )

    binning_group.add_argument(
        '--skip-binners', '--skip_binners', '--skip_binner', '--skip-binner',
        help='Optional list of binning algorithms to skip. Can be any combination of: \n'
             'rosella, semibin, metabat1, metabat2, metabat, vamb, concoct, maxbin2, maxbin \n'
             'Capitals will be auto-corrected. N.B. specifying "metabat" will skip both \n'
             'MetaBAT1 and MetaBAT2.',
        dest='skip_binners',
        nargs='*'
        # default=["maxbin2"]
    )

    ####################################################################
    mag_group = argparse.ArgumentParser(formatter_class=CustomHelpFormatter,
                                        add_help=False)
    mag_group_exclusive = mag_group.add_mutually_exclusive_group()

    # mag_group_exclusive.add_argument(
    #     '-f', '--genome-fasta-files', '--genome_fasta_files',
    #     help='MAGs to be annotated',
    #     dest='mags',
    #     nargs='*',
    #     required=False,
    # )

    mag_group_exclusive.add_argument(
        '-d', '--genome-fasta-directory', '--genome_fasta_directory',
        help='Directory containing MAGs to be annotated',
        dest='directory',
        required=False,
    )

    mag_group.add_argument(
        '-x', '--fasta-extension', '--fasta_extension',
        help='File extension of fasta files in --genome-fasta-directory',
        dest='ext',
        required=False,
        default='fna'
    )

    #####################################################################
    isolate_group = argparse.ArgumentParser(formatter_class=CustomHelpFormatter,
                                            add_help=False)

    isolate_group.add_argument(
        '--guppy-model', '--guppy_model',
        help='The guppy model used by medaka to perform polishing',
        dest='guppy_model',
        nargs=1,
        required=False,
        default='r941_min_high_g360'
    )

    isolate_group.add_argument(
        '--genome-size', '--genome_size',
        help='Approximate size of the isolate genome to be assembled',
        dest='genome_size',
        nargs=1,
        required=False,
        default=5000000
    )

    #####################################################################
    cluster_group = argparse.ArgumentParser(formatter_class=CustomHelpFormatter, add_help=False)

    cluster_group.add_argument(
        '--ani',
        help='Overall ANI level to dereplicate at with FastANI.',
        dest='ani',
        default=97
    )

    cluster_group.add_argument(
        '--precluster-ani', '--precluster_ani',
        help='Require at least this dashing-derived ANI for preclustering and to avoid FastANI on distant lineages within preclusters.',
        dest='precluster_ani',
        default=95
    )

    cluster_group.add_argument(
        '--precluster-method', '--precluster_method',
        help="method of calculating rough ANI for dereplication. 'dashing' for HyperLogLog, 'finch' for finch MinHash.",
        dest='precluster_method',
        default='dashing'
    )

    cluster_group.add_argument(
        '--min-completeness', '--min_completeness',
        help="Ignore genomes with less completeness than this percentage.",
        dest='min_completeness',
        default='none'
    )

    cluster_group.add_argument(
        '--max-contamination', '--max_contamination',
        help="Ignore genomes with more contamination than this percentage.",
        dest='max_contamination',
        default='none'
    )

    cluster_group.add_argument(
        '--use-checkm2-scores', '--use_checkm2_scores',
        help="Use CheckM2 completeness and contamination scores (if available) to perform Galah dereplication",
        type=str2bool,
        nargs='?',
        const=True,
        dest='use_checkm2_scores',
        default=False
    )

    cluster_group.add_argument(
        '--pggb-params', '--pggb_params',
        help="Parameters to be used with pggb, must be surrounded by quotation marks e.g. \'\'",
        dest='pggb_params',
        default='-k 79 -G 7919,8069'
    )

    #####################################################################
    # viral_group = argparse.ArgumentParser(formatter_class=CustomHelpFormatter,
    #                                         add_help=False)
    #
    # viral_group.add_argument(
    #     '--virsorter-data', '--virsorter_data',
    #     help='The guppy model used by medaka to perform polishing',
    #     dest='guppy_model',
    #     nargs=1,
    #     required=False,
    #     default='r941_min_high_g360'
    # )
    #
    # viral_group.add_argument(
    #     '--genome-size', '--genome_size',
    #     help='Approximate size of the isolate genome to be assembled',
    #     dest='genome_size',
    #     nargs=1,
    #     required=False,
    #     default=5000000
    # )

    assemble_group = argparse.ArgumentParser(formatter_class=CustomHelpFormatter, add_help=False)
    assemble_group.add_argument(
        '--use-unicycler', '--use_unicycler',
        help='Use Unicycler to re-assemble the metaSPAdes hybrid assembly. Not recommended for complex metagenomes.',
        type=str2bool,
        nargs='?',
        const=True,
        dest='use_unicycler',
        default=False,
    )

    assemble_group.add_argument(
        '--use-megahit', '--use_megahit',
        help='Specifies whether or not to use megahit if multiple for short-read only assembly',
        type=str2bool,
        nargs='?',
        const=True,
        dest='use_megahit',
        default=False,
    )

    assemble_group.add_argument(
        '--coassemble', '--co-assemble', '--co_assemble',
        help='Specifies whether or not, when given multiple input reads, to coassemble them. \n'
             'If False, Aviary will use the first set of short reads and first set of long reads to perform assembly \n'
             'All read files will still be used during the MAG recovery process for differential coverage.',
        type=str2bool,
        nargs='?',
        const=True,
        dest='coassemble',
        default=True,
    )

    assemble_group.add_argument(
        '--kmer-sizes', '--kmer_sizes', '-k',
        help='Manually specify the kmer-sizes used by SPAdes during assembly. Space separated odd integer values '
             'and less than 128 or "auto"',
        dest='kmer_sizes',
        default=['auto'],
        nargs="+",
    )

    #~#~#~#~#~#~#~#~#~#~#~#~#~   sub-parsers   ~#~#~#~#~#~#~#~#~#~#~#~#~#
    ##########################  ~ ASSEMBLE ~  ###########################

    assemble_options = subparsers.add_parser('assemble',
                                              description='Step-down hybrid assembly using long and short reads, or assembly using only short or long reads.',
                                              formatter_class=CustomHelpFormatter,
                                              parents=[qc_group, assemble_group, short_read_group, long_read_group, binning_group, base_group],
                                              epilog=
        '''
                                        ......:::::: ASSEMBLE ::::::......

        aviary assemble -1 *.1.fq.gz -2 *.2.fq.gz --longreads *.nanopore.fastq.gz --long_read_type ont

        ''')



    assemble_options.add_argument(
        '-w', '--workflow',
        help='Main workflow to run',
        dest='workflow',
        nargs="+",
        default=['complete_assembly'],
    )

    ##########################  ~ RECOVER ~   ###########################

    recover_options = subparsers.add_parser('recover',
                                            description='The aviary binning pipeline',
                                            formatter_class=CustomHelpFormatter,
                                            parents=[qc_group, assemble_group, short_read_group, long_read_group, binning_group, annotation_group, base_group],
                                            epilog=
    '''
                                           ......:::::: RECOVER ::::::......
    
    aviary recover --assembly scaffolds.fasta -1 *.1.fq.gz -2 *.2.fq.gz --longreads *.nanopore.fastq.gz --long_read_type ont

    ''')

    recover_options.add_argument(
        '-a', '--assembly',
        help='Optional FASTA file containing scaffolded contigs of the metagenome assembly',
        dest="assembly",
        nargs=1,
        required=False,
    )

    recover_options.add_argument(
        '-w', '--workflow',
        help='Main workflow to run',
        dest='workflow',
        nargs="+",
        default=['recover_mags'],
    )

    recover_options.add_argument(
        '--perform-strain-analysis', '--perform_strain_analysis',
        help='Specify whether to use Lorikeet on recovered MAGs get strain diversity metrics',
        type=str2bool,
        nargs='?',
        const=True,
        dest='strain_analysis',
        default=False
    )

    ##########################  ~ ANNOTATE ~   ###########################

    annotate_options = subparsers.add_parser('annotate',
                                              description='Annotate a given set of MAGs using EggNOG, GTDB-tk, and Checkm2',
                                              formatter_class=CustomHelpFormatter,
                                              parents=[mag_group, annotation_group, base_group, qc_group],
                                              epilog=
                                            '''
                                                  ......:::::: ANNOTATE ::::::......
                                        
                                            aviary annotate --genome-fasta-directory input_bins/
                                        
                                            ''')

    annotate_options.add_argument(
        '-a', '--assembly',
        help='FASTA file containing scaffolded contigs of one or more metagenome assemblies wishing to be passed to QUAST',
        dest="assembly",
        nargs="*",
        required=False,
    )

    annotate_options.add_argument(
        '-w', '--workflow',
        help='Main workflow to run',
        dest='workflow',
        nargs="+",
        default=['annotate'],
    )

    ##########################  ~ diversity ~   ###########################

    diversity_options = subparsers.add_parser('diversity',
                                             description='Perform strain diversity analysis',
                                             formatter_class=CustomHelpFormatter,
                                             parents=[mag_group, qc_group, assemble_group, short_read_group, long_read_group,
                                                      binning_group, annotation_group, base_group],
                                             epilog=
                                             '''
                                                                    ......:::::: DIVERSITY ::::::......

                                             aviary diversity -c R1.fastq.gz R2.fastq.gz --genome-fasta-directory input_bins/

                                             ''')

    diversity_options.add_argument(
        '-a', '--assembly',
        help='FASTA file containing scaffolded contigs of one or more metagenome assemblies wishing to be passed to QUAST',
        dest="assembly",
        nargs="*",
        required=False,
    )

    diversity_options.add_argument(
        '-w', '--workflow',
        help='Main workflow to run',
        dest='workflow',
        nargs="+",
        default=['lorikeet'],
    )

    diversity_options.add_argument(
        '--perform-strain-analysis', '--perform_strain_analysis',
        help=argparse.SUPPRESS,
        type=str2bool,
        nargs='?',
        const=True,
        dest='strain_analysis',
        default=True
    )

    ##########################  ~ CLUSTER ~   ###########################

    cluster_options = subparsers.add_parser('cluster',
                                             description='Clusters previous aviary runs together and performs'
                                                         'dereplication using Galah',
                                             formatter_class=CustomHelpFormatter,
                                             parents=[base_group, cluster_group],
                                             epilog=
                                             '''
                                                                   ......:::::: CLUSTER ::::::......

                                             aviary cluster --input-runs aviary_output_folder_1/ aviary_output_folder_2/

                                             ''')

    cluster_options.add_argument(
        '-i', '--input-runs', '--input_runs',
        help='The paths to the previous finished runs of Aviary. Must contain the bins/checkm.out and bins/final_bins'
             'outputs',
        dest='previous_runs',
        nargs='*',
        required=True,
    )

    cluster_options.add_argument(
        '-w', '--workflow',
        help='Main workflow to run',
        dest='workflow',
        nargs="+",
        default=['complete_cluster'],
    )

    ##########################  ~ VIRAL ~   ###########################

    viral_options = subparsers.add_parser('viral',
                                          description='The incomplete binning pipeline',
                                          formatter_class=CustomHelpFormatter,
                                          parents=[mag_group, short_read_group, long_read_group, annotation_group, base_group],
                                          epilog=
                                          '''
                                                  ......:::::: VIRAL ::::::...... 
 
                                          aviary viral --genome-fasta-files *.fasta
 
                                          ''')

    viral_options.add_argument(
        '-w', '--workflow',
        help='Main workflow to run',
        dest='workflow',
        nargs="+",
        default=['create_webpage_genotype'],
    )

    ##########################   ~ COMPLETE ~  ###########################

    complete_options = subparsers.add_parser('complete',
                                            description='Performs all steps in the Aviary pipeline. '
                                                        'Assembly > Binning > Refinement > Annotation > Diversity',
                                            formatter_class=CustomHelpFormatter,
                                            parents=[qc_group, assemble_group, short_read_group, long_read_group, binning_group, annotation_group, base_group],
                                            epilog=
                                            '''
                                                               ......:::::: COMPLETE ::::::......

                                            aviary complete -1 *.1.fq.gz -2 *.2.fq.gz --longreads *.nanopore.fastq.gz 

                                            ''')

    complete_options.add_argument(
        '-a', '--assembly',
        help='Optional FASTA file containing scaffolded contigs of the metagenome assembly',
        dest="assembly",
        nargs=1,
        required=False,
    )

    complete_options.add_argument(
        '-w', '--workflow',
        help='Main workflow to run',
        dest='workflow',
        nargs="+",
        default=['get_bam_indices', 'recover_mags', 'annotate', 'lorikeet'],
    )

    ##########################  ~ ISOLATE ~  ###########################

    isolate_options = subparsers.add_parser('isolate',
                                             description='Step-down hybrid assembly using long and short reads, or assembly using only short or long reads.',
                                             formatter_class=CustomHelpFormatter,
                                             parents=[qc_group, short_read_group, long_read_group, isolate_group, binning_group, annotation_group, base_group],
                                             epilog=
                                             '''
                                                                             ......:::::: ISOLATE ::::::......
                                 
                                             aviary isolate -1 *.1.fq.gz -2 *.2.fq.gz --longreads *.nanopore.fastq.gz --long_read_type ont
                                 
                                             ''')

    isolate_options.add_argument(
        '-w', '--workflow',
        help='Main workflows to run',
        dest='workflow',
        nargs="+",
        default=['circlator'],
    )

    ##########################   ~ BATCH ~  ###########################

    batch_options = subparsers.add_parser('batch',
                                             description='Performs all steps in the Aviary pipeline on a batch file. \n'
                                                         'Each line in the batch file is processed separately and then \n'
                                                         'clustered using aviary. \n'
                                                         '(Assembly > Binning > Refinement > Annotation > Diversity) * n_samples --> Cluster',
                                             formatter_class=CustomHelpFormatter,
                                             parents=[qc_group, assemble_group, binning_group, annotation_group, cluster_group, base_group],
                                             epilog=
                                             '''
                                                      ......:::::: BATCH ::::::......

                                             aviary batch -f batch_file.tsv -t 32 -o batch_test
                                             
                                             An example batch file can be found at: 

                                             ''')

    batch_options.add_argument(
        '-f', '--batch_file', '--batch-file',
        help='The tab or comma separated batch file containing the input samples to assemble and/or recover MAGs from. \n'
             'An example batch file can be found at XXX. The heading line is required. \n'
             'The number of reads provided to each sample is flexible as is the type of assembly being performed (if any). \n'
             'Multiple reads can be supplied by providing a comma-separated list (surrounded by double quotes \"\" if using a \n'
             'comma separated batch file) within the specific read column.',
        dest="batch_file",
        # nargs=1,
        required=True,
    )

    batch_options.add_argument(
        '--write-script', '--write_script',
        help='Write the aviary batch Snakemake commands to a bash script and exit. \n'
             'Useful when submitting jobs to HPC cluster with custom queueing.',
        dest='write_script',
        required=False
    )

    batch_options.add_argument(
        '--cluster',
        help='Cluster final output of all samples using aviary cluster if possible.',
        dest='cluster',
        type=str2bool,
        nargs='?',
        const=True,
        default=True
    )

    batch_options.add_argument(
        '--cluster-ani-values', '--cluster_ani_values', '--ani-values', '--ani_values',
        help='The range of ANI values to perform clustering and dereplication at during aviary cluster.',
        dest='ani_values',
        nargs='*',
        default=[0.99, 0.97, 0.95]
    )

    batch_options.add_argument(
        '--min-percent-read-identity-long', '--min_percent_read_identity_long',
        help='Minimum percent read identity used by CoverM for long-reads'
             'when calculating genome abundances.',
        dest='long_percent_identity',
        default='85'
    )

    batch_options.add_argument(
        '--min-percent-read-identity-short', '--min_percent_read_identity_short',
        help='Minimum percent read identity used by CoverM for short-reads \n'
             'when calculating genome abundances.',
        dest='short_percent_identity',
        default='95'
    )

    batch_options.add_argument(
        '-w', '--workflow',
        help='Main workflow to run for each sample',
        dest='workflow',
        nargs="+",
        default=['get_bam_indices', 'recover_mags', 'annotate', 'lorikeet'],
    )

    ##########################   ~ configure ~  ###########################

    configure_options = subparsers.add_parser('configure',
                                            description='Sets the conda environment variables for future runs and downloads databases. ',
                                            formatter_class=CustomHelpFormatter,
                                            parents=[base_group],
                                            epilog=
                                            '''
                                                               ......:::::: CONFIGURE ::::::......

                                            aviary configure --conda-prefix ~/.conda --gtdb-path ~/gtdbtk/release207/ --temp-dir /path/to/new/temp

                                            ''')

    configure_options.add_argument(
        '--gtdb-path', '--gtdb_path',
        help='Path to the local gtdb database files',
        dest='gtdb_path',
        required=False,
    )

    configure_options.add_argument(
        '--busco-db-path', '--busco_db_path',
        help='Path to the local BUSCO database files',
        dest='busco_db_path',
        required=False,
    )

    configure_options.add_argument(
        '--checkm2-db-path', '--checkm2_db_path',
        help='Path to Checkm2 Database',
        dest='checkm2_db_path',
        required=False,
    )

    configure_options.add_argument(
        '--eggnog-db-path', '--eggnog_db_path',
        help='Path to the local eggnog database files',
        dest='eggnog_db_path',
        required=False,
    )

    configure_options.add_argument(
        '-w', '--workflow',
        help=argparse.SUPPRESS,
        dest='workflow',
        nargs="+",
        default=['download_databases'],
    )

    ###########################################################################
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Parsing input ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    args = main_parser.parse_args()
    time = datetime.now().strftime('%H:%M:%S %d-%m-%Y')

    if args.log:
        if os.path.isfile(args.log):
            raise Exception("File %s exists" % args.log)
        logging.basicConfig(filename=args.log,
                            level=debug[args.verbosity],
                            format='%(asctime)s %(levelname)s: %(message)s',
                            datefmt='%m/%d/%Y %I:%M:%S %p')
    else:
        logging.basicConfig(level=debug[args.verbosity],
                            format='%(asctime)s %(levelname)s: %(message)s',
                            datefmt='%m/%d/%Y %I:%M:%S %p')
    logging.info("Time - %s" % (time))
    logging.info("Command - %s" % ' '.join(sys.argv))
    logging.info("Version - %s" % __version__)

    if args.subparser_name == 'configure':
        # Set the environment variables if manually configuring
        if args.conda_prefix is not None:
            Config.set_db_path(args.conda_prefix, db_name='CONDA_ENV_PATH')

        if args.tmpdir is not None:
            Config.set_db_path(args.tmpdir, db_name='TMPDIR')

        if args.gtdb_path is not None:
            Config.set_db_path(args.gtdb_path, db_name='GTDBTK_DATA_PATH')

        if args.busco_db_path is not None:
            Config.set_db_path(args.busco_db_path, db_name='BUSCO_DB')

        if args.checkm2_db_path is not None:
            Config.set_db_path(args.checkm2_db_path, db_name='CHECKM2DB')

        if args.eggnog_db_path is not None:
            Config.set_db_path(args.eggnog_db_path, db_name='EGGNOG_DATA_DIR')

        logging.info("The current aviary environment variables are:")
        logging.info(f"CONDA_ENV_PATH: {Config.get_software_db_path('CONDA_ENV_PATH', '--conda-prefix')}")
        logging.info(f"TMPDIR: {Config.get_software_db_path('TMPDIR', '--tmpdir')}")
        logging.info(f"GTDBTK_DATA_PATH: {Config.get_software_db_path('GTDBTK_DATA_PATH', '--gtdb-path')}")
        logging.info(f"EGGNOG_DATA_DIR: {Config.get_software_db_path('EGGNOG_DATA_DIR', '--eggnog-db-path')}")
        logging.info(f"CHECKM2DB: {Config.get_software_db_path('CHECKM2DB', '--checkm2-db-path')}")
        if not args.download:
            logging.info("All paths set. Exitting without downloading databases. If you wish to download databases use --download")
            sys.exit(0)

    # else:
    args = manage_env_vars(args)
    prefix = args.output
    if not os.path.exists(prefix):
        os.makedirs(prefix)

    if args.subparser_name != 'batch':
        processor = Processor(args)
        processor.make_config()

        if args.build:
            try:
                args.cmds = args.cmds + '--conda-create-envs-only '
            except TypeError:
                args.cmds = '--conda-create-envs-only '

        try:
            if args.subparser_name == 'assemble':
                if args.use_unicycler:
                    args.workflow.insert(0, "combine_assemblies")
        except AttributeError:
            pass

        processor.run_workflow(cores=int(args.n_cores),
                               dryrun=args.dryrun,
                               clean=args.clean,
                               conda_frontend=args.conda_frontend,
                               snakemake_args=args.cmds)
    else:
        process_batch(args, prefix)

def manage_env_vars(args):
    if args.conda_prefix is None:
        args.conda_prefix = Config.get_software_db_path('CONDA_ENV_PATH', '--conda-prefix')

    if args.tmpdir is None:
        args.tmpdir = Config.get_software_db_path('TMPDIR', '--tmpdir')

    try:
        if args.gtdb_path is None:
            args.gtdb_path = Config.get_software_db_path('GTDBTK_DATA_PATH', '--gtdb-path')
        if args.eggnog_db_path is None:
            args.eggnog_db_path = Config.get_software_db_path('EGGNOG_DATA_DIR', '--eggnog-db-path')
        if args.checkm2_db_path is None:
            args.checkm2_db_path = Config.get_software_db_path('CHECKM2DB', '--checkm2-db-path')
    except AttributeError:
        pass

    return args

###############################################################################
################################ - Classes - ##################################

class CustomHelpFormatter(argparse.HelpFormatter):
    def _split_lines(self, text, width):
        return text.splitlines()

    def _get_help_string(self, action):
        h = action.help
        if '%(default)' not in action.help:
            if action.default != '' and \
               action.default != [] and \
               action.default != None \
               and action.default != False:
                if action.default is not argparse.SUPPRESS:
                    defaulting_nargs = [argparse.OPTIONAL,
                                        argparse.ZERO_OR_MORE]

                    if action.option_strings or action.nargs in defaulting_nargs:

                        if '\n' in h:
                            lines = h.splitlines()
                            lines[0] += ' (default: %(default)s)'
                            h = '\n'.join(lines)
                        else:
                            h += ' (default: %(default)s)'
        return h

    def _fill_text(self, text, width, indent):
        return ''.join([indent + line for line in text.splitlines(True)])

if __name__ == '__main__':
    sys.exit(main())


import pysam
import os

contig_bins = {}
outlength = {}
samfile = pysam.AlignmentFile(snakemake.input.long_bam, 'rb')
gotten_contigs = set()
maxbin = 0
outreads = {}
outbases = {}
outreads500 = {}
outbases500 = {}

for bins in os.listdir(snakemake.input.metabat_done[:-4]):
    if not bins.startswith(
            'binned_contigs') or bins == "binned_contigs.unbinned" or bins == "binned_contigs.lowDepth" or bins == "binned_contigs.tooShort":
        continue
    bin = bins.split('.')[1]
    if int(bin) > maxbin:
        maxbin = int(bin)
    outlength[bin] = 0
    outreads[bin] = set()
    outbases[bin] = 0
    outreads500[bin] = set()
    outbases500[bin] = 0
    with open(os.path.join(snakemake.input.metabat_done[:-4], bins)) as f:
        for line in f:
            contig_bins[line.rstrip()] = bin
            gotten_contigs.add(line.rstrip())
            outlength[bin] += samfile.get_reference_length(line.rstrip())

cutoff = 0.05

for read in samfile.fetch(until_eof=True):
    start = True
    if not read.cigartuples is None:
        clipped_start = 0
        clipped_end = 0
        for i in read.cigartuples:
            if i[0] == 4 or i[0] == 5:
                if start:
                    clipped_start += i[1]
                else:
                    clipped_end += i[1]
            else:
                start = False
        if read.reference_name in contig_bins:
            bin = contig_bins[read.reference_name]
            length = read.infer_read_length()
            if (clipped_start / length <= cutoff and clipped_end / length <= cutoff) or \
                    (clipped_start / length <= cutoff and read.reference_end > read.reference_length - 100) or \
                    (clipped_end / length <= cutoff and read.reference_start < 100) or \
                    (read.reference_start < 100 and read.reference_end > read.reference_length - 100):
                outreads[bin].add(read.query_name)
                outbases[bin] += length

try:
    os.makedirs("data/binned_reads")
except FileExistsError:
    pass

out_dict = {}
for i in outreads:
    with open("data/binned_reads/r" + i + '.long.list', 'w') as read_list:
        reads = outreads[i]
        bases = outbases[i]
        for j in reads:
            read_list.write(j + '\n')
    out_dict[i] = [i, "data/binned_reads/r" + i + '.long.list', str(outlength[i]), str(bases)]

samfile = pysam.AlignmentFile(snakemake.input.short_bam, 'rb')
outreads = {}
outbases = {}
for read in samfile.fetch(until_eof=True):
    start = True
    if read.is_proper_pair:
        if read.reference_name in contig_bins:
            bin = contig_bins[read.reference_name]
            length = read.infer_read_length()
            if not bin in outreads:
                outreads[bin] = set()
                outbases[bin] = 0
            outreads[bin].add(read.query_name)
            outbases[bin] += length
for i in outreads:
    with open("data/binned_reads/r" + i + '.short.list', 'w') as read_list:
        for j in outreads[i]:
            if reads == 'data/short_reads.fastq.gz' or snakemake.config['short_reads_2'] == 'none':
                if j[:-2] in ['/1', '/2']:
                    read_list.write(j[:-2] + '/1\n')
                    read_list.write(j[:-2] + '/2\n')
                else:
                    read_list.write(j + '/1\n')
                    read_list.write(j + '/2\n')
            else:
                read_list.write(j + '\n')
    out_dict[i] += ["data/binned_reads/r" + i + '.short.list', str(outbases[i])]

with open(snakemake.output.list, 'w') as o:
    for i in sorted(list(out_dict), key=int):
        o.write('\t'.join(out_dict[i]) + '\n')



import io
from os.path import dirname, join
from setuptools import setup, find_packages


# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


def get_version(relpath):
  """Read version info from a file without importing it"""
  for line in io.open(join(dirname(__file__), relpath), encoding="cp437"):
    if "__version__" in line:
      if '"' in line:
        # __version__ = "0.9"
        return line.split('"')[1]
      elif "'" in line:
        return line.split("'")[1]


setup(
    name='aviary-genome',
    version=get_version("aviary/__init__.py"),
    url='https://github.com/rhysnewell/aviary',
    license='GPL-3.0',
    author='Rhys Newell',
    author_email='rhys.newell94@gmail.com',
    description='aviary - metagenomics pipeline using long and short reads',
    long_description=long_description,
    long_description_content_type='text/markdown',
    zip_safe=False,
    packages=find_packages(),
    package_data={'template_config': ['aviary/modules/template_config.yaml'],
                  'Snakefile': ['aviary/modules/Snakefile']},
    data_files=[(".", ["README.md", "LICENSE"])],
    include_package_data=True,
    install_requires= [
        "snakemake",
        "numpy",
        "ruamel.yaml>=0.15.99",
        "pandas",
        "biopython",
    ],
    entry_points={
          'console_scripts': [
              'aviary = aviary.aviary:main'
          ]
    },
    classifiers=["Topic :: Scientific/Engineering :: Bio-Informatics"],
)



import pandas as pd
import subprocess
import shutil
import glob
import os
import pandas.errors as e

def refinery():
    """
    Main function that performs the refining process iteratively passing bins between
    rosella refine and checkm until either the max number of iterations is reached
    or there are no more contaminated bins
    """
    # These will change
    current_iteration = 0
    checkm_path = snakemake.input.checkm

    # These will not change
    coverage = snakemake.input.coverage
    assembly = snakemake.input.fasta
    final_refining = snakemake.params.final_refining

    # kmers = snakemake.input.kmers
    if os.path.isfile("data/rosella_bins/rosella_kmer_table.tsv"):
        kmers = "data/rosella_bins/rosella_kmer_table.tsv"
    else:
        kmers = None
    min_bin_size = snakemake.params.min_bin_size
    max_iterations = int(snakemake.params.max_iterations)
    pplacer_threads = int(snakemake.params.pplacer_threads)
    threads = int(snakemake.threads)
    max_contamination = int(snakemake.params.max_contamination)
    contaminated_bin_folder = snakemake.params.output_folder + "/contaminated_bins"
    final_bins = snakemake.params.output_folder + "/final_bins"

    try:
        current_checkm = pd.read_csv(checkm_path, sep='\t', comment="[")
    except e.EmptyDataError:
        open(f"{snakemake.params.output_folder}/done", "a").close()
        return


    bin_folder = snakemake.params.bin_folder
    extension = snakemake.params.extension


    try:
        os.makedirs(contaminated_bin_folder)
    except FileExistsError:
        shutil.rmtree(contaminated_bin_folder)
        os.makedirs(contaminated_bin_folder)

    try:
        os.makedirs(final_bins)
    except FileExistsError:
        shutil.rmtree(final_bins)
        os.makedirs(final_bins)

    final_checkm = current_checkm.copy().loc[current_checkm["Contamination"] <= snakemake.params.max_contamination].copy()
    final_checkm = move_finished_bins(final_checkm, bin_folder, extension, final_bins)


    while current_iteration < max_iterations:

        # delete previous contaminated set
        if current_iteration != 0:
            files = glob.glob(f'{contaminated_bin_folder}/*')
            [os.remove(file) for file in files]
            extension = "fna"

        # put contaminated bins in one folder. If no contaminated bins then break
        if not collect_contaminated_bins(
                current_checkm, max_contamination, bin_folder, extension, contaminated_bin_folder):
            break

        output_folder = f"{snakemake.params.output_folder}/rosella_refined_{current_iteration}"

        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)
        # Refine the contaminated bins
        kmers = refine(assembly, coverage, kmers, checkm_path,
                   contaminated_bin_folder, extension, min_bin_size,
                   threads, output_folder, max_contamination)

        # update the bin folder variable to the current refined folder
        bin_folder = output_folder

        # retrieve the checkm results for the refined bins
        try:
            get_checkm_results(bin_folder, threads, pplacer_threads, final_refining)
            # update the checkm results and counter
            checkm_path = f"{bin_folder}/checkm.out"
            current_checkm = pd.read_csv(checkm_path, sep='\t', comment='[')

            bins_to_keep = current_checkm.copy().loc[current_checkm["Contamination"] <= max_contamination]
            bins_to_keep = move_finished_bins(bins_to_keep, bin_folder, "fna", final_bins, current_iteration)
            final_checkm = pd.concat([final_checkm, bins_to_keep])
            current_iteration += 1
        except FileNotFoundError:
            # No bins to refine, break out and move on
            break

    if final_refining:
        final_checkm.to_csv("data/checkm.out", sep='\t', index=False)
        final_output_folder = "bins/final_bins"
        if os.path.exists("bins/"): # remove pre-existing output, preventing bin duplication
            shutil.rmtree("bins/")
        os.makedirs("bins/")
        if not os.path.exists(final_output_folder):
            os.symlink("../" + final_bins, final_output_folder, target_is_directory=True)
        elif not os.path.islink(final_output_folder):
            os.rmdir(final_output_folder)
            os.symlink("../" + final_bins, final_output_folder, target_is_directory=True)
        final_checkm.to_csv("bins/checkm.out", sep='\t', index=False)
    else:
        open(f"{snakemake.params.output_folder}/done", "a").close()


def move_finished_bins(
        checkm,
        input_folder,
        input_extension,
        output_folder,
        current_iteration = None
):
    try:
        bins = checkm["Bin Id"]
        column_name = "Bin Id"
    except KeyError:
        # checkm2 input
        bins = checkm["Name"]
        column_name = "Name"

    new_bin_ids = []
    for bin_id in bins:
        # print(f"{input_folder}/{bin_id}")
        if current_iteration is not None:
            new_bin_id = f"{bin_id}_{current_iteration}"
        else:
            new_bin_id = bin_id

        new_bin_ids.append(new_bin_id)
        try:
            shutil.copy(f"{input_folder}/{bin_id}.{input_extension}", f"{output_folder}/{new_bin_id}.fna")
        except FileNotFoundError:
            # File already moved
            continue

    checkm[column_name] = new_bin_ids

    return checkm


def collect_contaminated_bins(
        checkm,
        max_contamination,
        bin_folder,
        extension,
        output_folder
):
    contaminated = checkm[checkm["Contamination"] > max_contamination]

    # check that there is at least one contaminated bin
    if contaminated.shape[0] == 0: return False

    try:
        bins = contaminated["Bin Id"]
    except KeyError:
        # checkm2 input
        bins = contaminated["Name"]


    for bin_id in bins:
        try:
            shutil.copy(f"{bin_folder}/{bin_id}.{extension}", f"{output_folder}/{bin_id}.{extension}")
        except FileNotFoundError:
            # file already moved
            continue

    return True


def refine(
        assembly, coverage, kmers, checkm,
        bin_folder, extension, min_bin_size,
        threads, output_folder, max_contamination,
):
    if kmers is None:
        subprocess.Popen(f"rosella refine -a {assembly} --coverage-values {coverage} "
                         f"-d {bin_folder} -x {extension} --checkm-file {checkm} --max-contamination {max_contamination} "
                         f"--min-bin-size {min_bin_size} -t {threads} -o {output_folder}", shell=True).wait()
        os.makedirs("data/rosella_bins/", exist_ok=True)
        shutil.copy(f"{output_folder}/rosella_kmer_table.tsv", f"data/rosella_bins/rosella_kmer_table.tsv")
        kmers = "data/rosella_bins/rosella_kmer_table.tsv"
    else:
        subprocess.Popen(f"rosella refine -a {assembly} --coverage-values {coverage} --kmer-frequencies {kmers} "
                         f"-d {bin_folder} -x {extension} --checkm-file {checkm} --max-contamination {max_contamination} "
                         f"--min-bin-size {min_bin_size} -t {threads} -o {output_folder}", shell=True).wait()

    return kmers

def get_checkm_results(
        refined_folder,
        threads,
        pplacer_threads,
        final_refining=False
):

    subprocess.Popen(f"checkm lineage_wf -t {threads} --pplacer_threads {pplacer_threads} -x fna "
                     f"--tab_table -f {refined_folder}/checkm.out {refined_folder} {refined_folder}/checkm",
                     shell=True).wait()

    if final_refining:
        subprocess.Popen(f"checkm qa -o 2 --tab_table -f {refined_folder}/checkm.out "
                         f"{refined_folder}/checkm/lineage.ms {refined_folder}/checkm/",
                         shell=True).wait()


if __name__ == '__main__':
    refinery()


import subprocess
import os
import logging

logging.info("Generating BAM files for pilon...")
if os.path.exists('data/short_reads.fastq.gz'):
    subprocess.Popen("minimap2 -ax sr -t %d %s data/short_reads.fastq.gz | samtools view -b -F 4 -@ %d | "
                     "samtools sort -@ %d -o %s -; samtools index -@ %d %s; " %
                     (snakemake.threads, snakemake.input.fasta,
                      snakemake.threads, snakemake.threads, snakemake.output.bam,
                      snakemake.threads, snakemake.output.bam,
                      ), shell=True).wait()

elif snakemake.config['short_reads_2'] != 'none':
    if len(snakemake.config['short_reads_2']) == 1 or not snakemake.params.coassemble:
        pe1 = snakemake.config['short_reads_1'][0]
        pe2 = snakemake.config['short_reads_2'][0]
    else:
        if not os.path.exists("data/short_reads.1.fastq.gz"):
            for reads1, reads2 in zip(snakemake.config['short_reads_1'], snakemake.config['short_reads_2']):
                subprocess.Popen(f"cat {reads1} >> data/short_reads.1.fastq.gz", shell=True).wait()
                subprocess.Popen(f"cat {reads2} >> data/short_reads.2.fastq.gz", shell=True).wait()
        pe1 = "data/short_reads.1.fastq.gz"
        pe2 = "data/short_reads.2.fastq.gz"

    subprocess.Popen("minimap2 -ax sr -t %d %s %s %s | samtools view -b -F 4 -@ %d | "
                     "samtools sort -@ %d -o %s -; samtools index -@ %d %s; " %
                     (snakemake.threads, snakemake.input.fasta, pe1, pe2,
                      snakemake.threads, snakemake.threads, snakemake.output.bam,
                      snakemake.threads, snakemake.output.bam,
                      ), shell=True).wait()

    # os.remove("data/short_reads.1.fastq.gz")
    # os.remove("data/short_reads.2.fastq.gz")

elif snakemake.config['short_reads_1']  != 'none':
    if len(snakemake.config['short_reads_1']) == 1:
        pe1 = snakemake.config['short_reads_1'][0]
    else:
        if not os.path.exists("data/short_reads.1.fastq.gz") or not snakemake.params.coassemble:
            for reads1 in snakemake.config['short_reads_1']:
                subprocess.Popen(f"cat {reads1} >> data/short_reads.1.fastq.gz", shell=True).wait()
        pe1 = "data/short_reads.1.fastq.gz"

    subprocess.Popen("minimap2 -ax sr -t %d %s %s | samtools view -b -F 4 -@ %d | "
                     "samtools sort -@ %d -o %s -; samtools index -@ %d %s; " %
                     (snakemake.threads, snakemake.input.fasta, pe1,
                      snakemake.threads, snakemake.threads, snakemake.output.bam,
                      snakemake.threads, snakemake.output.bam,
                      ), shell=True).wait()

    # os.remove("data/short_reads.1.fastq.gz")



import os
import subprocess
import random
import shutil
import logging

out = "data/racon_polishing"
max_cov = snakemake.params.maxcov

try:
    os.makedirs(out)
except FileExistsError:
    pass

random.seed(89)
reference = snakemake.input.fasta

# Whether contigs are polished with illumina or long read
if snakemake.params.illumina:
    if snakemake.config['reference_filter'] != 'none':
        reads = "data/short_reads.fastq.gz"
    elif snakemake.config['short_reads_2'] != 'none':
        # if len(snakemake.config['short_reads_2']) == 1 or not snakemake.params.coassemble:
        #     pe1 = snakemake.config['short_reads_1'][0]
        #     pe2 = snakemake.config['short_reads_2'][0]
        # else:
        # Racon can't handle paired end reads. It treats them as singled-ended. But when you have paired end reads
        # in separate files they can share the same read name, so we need to alter the read name based on the pair
        if not os.path.exists("data/short_reads.racon.1.fastq.gz"):
            for reads1, reads2 in zip(snakemake.config['short_reads_1'], snakemake.config['short_reads_2']):
                cat_or_zcat1 = "cat"
                cat_or_zcat2 = "cat"
                if reads1[-3::] == ".gz":
                    cat_or_zcat1 = "zcat"
                if reads2[-3::] == ".gz":
                    cat_or_zcat2 = "zcat"

                subprocess.Popen(f"{cat_or_zcat1} {reads1} | sed 's/@/@1_/' >> data/short_reads.racon.1.fastq", shell=True).wait()
                subprocess.Popen(f"{cat_or_zcat2} {reads2} | sed 's/@/@2_/' >> data/short_reads.racon.1.fastq", shell=True).wait()
                if not snakemake.params.coassemble:
                    break

            subprocess.Popen(f"pigz -p {snakemake.threads} --fast data/short_reads.racon.1.fastq", shell=True).wait()
            # subprocess.Popen(f"pigz -p {snakemake.threads} --fast data/short_reads.racon.2.fastq", shell=True).wait()

        pe1 = "data/short_reads.racon.1.fastq.gz"
        # pe2 = "data/short_reads.racon.2.fastq.gz"
        reads = [pe1]
    else:
        if len(snakemake.config['short_reads_1']) == 1 or not snakemake.params.coassemble:
            pe1 = snakemake.config['short_reads_1'][0]
        else:
            if not os.path.exists("data/short_reads.1.fastq.gz"):
                for reads1 in snakemake.config['short_reads_1']:
                    cat_or_zcat1 = "cat"
                    if reads1[-3::] == ".gz":
                        cat_or_zcat1 = "zcat"
                    subprocess.Popen(f"{cat_or_zcat1} {reads1} >> data/short_reads.1.fastq", shell=True).wait()

                subprocess.Popen(f"pigz -p {snakemake.threads} --fast data/short_reads.1.fastq",
                                 shell=True).wait()
            pe1 = "data/short_reads.1.fastq.gz"
        reads = [pe1]
else:
    reads = snakemake.input.fastq

for rounds in range(snakemake.params.rounds):
    paf = os.path.join(out, 'alignment.%s.%d.paf') % (snakemake.params.prefix, rounds)
    print("Generating PAF file: %s for racon round %d..." % (paf, rounds))

    # Generate PAF mapping files
    if not os.path.exists(paf): # Check if mapping already exists
        if snakemake.params.illumina:
            if reads != "data/short_reads.fastq.gz":
                subprocess.Popen("minimap2 -t %d -x sr %s %s > %s" % (snakemake.threads, reference, ' '.join(reads), paf),
                                 shell=True).wait()
            else:
                subprocess.Popen("minimap2 -t %d -x sr %s %s > %s" % (snakemake.threads, reference, reads, paf),
                                 shell=True).wait()
        elif snakemake.config["long_read_type"] in ['ont', 'ont_hq']:
            subprocess.Popen("minimap2 -t %d -x map-ont %s %s > %s" % (snakemake.threads, reference, reads, paf), shell=True).wait()
        else:
            subprocess.Popen("minimap2 -t %d -x map-pb %s %s > %s" % (snakemake.threads, reference, reads, paf), shell=True).wait()

    cov_dict = {}
    # Populate coverage dictionary,
    with open(paf) as f:
        for line in f:
            qname, qlen, qstart, qstop, strand, ref, rlen, rstart, rstop = line.split()[:9]
            qlen, qstart, qstop, rlen, rstart, rstop = map(int, [qlen, qstart, qstop, rlen, rstart, rstop])
            if ref in cov_dict:
                cov_dict[ref] += (rstop - rstart) / rlen
            else:
                cov_dict[ref] = (rstop - rstart) / rlen

    high_cov = set()
    low_cov = set()
    for i in cov_dict:
        if cov_dict[i] >= max_cov:
            high_cov.add(i)
        else:
            low_cov.add(i)

    no_cov = set()
    with open(reference) as ref_file, open(os.path.join(out, "filtered.%s.%d.fa" % (snakemake.params.prefix, rounds)), 'w') as o:
        for line in ref_file:
            if line.startswith('>'):
                name = line.split()[0][1:]
                if name in low_cov or name in high_cov:
                    o.write(line)
                    getseq = True
                else:
                    no_cov.add(name)
                    getseq = False
            elif getseq:
                o.write(line)

    included_reads = set()
    excluded_reads = set()
    with open(paf) as f, open(os.path.join(out, "filtered.%s.%d.paf" % (snakemake.params.prefix, rounds)), 'w') as paf_file:
        for line in f:
            qname, qlen, qstart, qstop, strand, ref, rlen, rstart, rstop = line.split()[:9]
            qlen, qstart, qstop, rlen, rstart, rstop = map(int, [qlen, qstart, qstop, rlen, rstart, rstop])
            if snakemake.params.illumina:
                if qname[:-2] in ['/1', '/2']:
                    qname = qname[:-2]
            if ref in low_cov:
                paf_file.write(line)
                included_reads.add(qname)
            elif ref in high_cov:
                # Down sample reads from high coverage contigs
                sample_rate = max_cov / cov_dict[ref]
                if qname in excluded_reads:
                    pass
                elif qname in included_reads:
                    paf_file.write(line)
                elif random.random() < sample_rate:
                    included_reads.add(qname)
                    paf_file.write(line)
                else:
                    excluded_reads.add(qname)
    with open(os.path.join(out, "reads.%s.%d.lst" % (snakemake.params.prefix, rounds)), "w") as o:
        for i in included_reads:
            if (reads == 'data/short_reads.fastq.gz' or snakemake.config['short_reads_2'] == 'none') and snakemake.params.illumina:
                o.write(i + '/1\n')
                o.write(i + '/2\n')
            else:
                o.write(i + '\n')
    logging.info("Retrieving reads...")
    if not isinstance(reads, str):
        for read in reads:
            seqkit_command = f"seqkit -j {snakemake.threads} grep --pattern-file {out}/reads.{snakemake.params.prefix}.{rounds}.lst {read} | pigz -p {snakemake.threads} >> {out}/reads.{snakemake.params.prefix}.{rounds}.fastq.gz"
            print(seqkit_command)
            subprocess.Popen(seqkit_command, shell=True).wait()
    else:
        seqkit_command = f"seqkit -j {snakemake.threads} grep --pattern-file {out}/reads.{snakemake.params.prefix}.{rounds}.lst {reads} | pigz -p {snakemake.threads} >> {out}/reads.{snakemake.params.prefix}.{rounds}.fastq.gz"
        print(seqkit_command)
        subprocess.Popen(seqkit_command, shell=True).wait()

    print("Performing round %d of racon polishing..." % rounds)
    print("racon -m 8 -x -6 -g -8 -w 500 -t %d -u %s/reads.%s.%d.fastq.gz %s/filtered.%s.%d.paf %s/filtered.%s.%d.fa"
                     " > %s/filtered.%s.%d.pol.fa" % (snakemake.threads, out, snakemake.params.prefix, rounds, out,
                                                   snakemake.params.prefix, rounds, out, snakemake.params.prefix, rounds,
                                                   out, snakemake.params.prefix, rounds))
    subprocess.Popen("racon -m 8 -x -6 -g -8 -w 500 -t %d -u %s/reads.%s.%d.fastq.gz %s/filtered.%s.%d.paf %s/filtered.%s.%d.fa"
                     " > %s/filtered.%s.%d.pol.fa" % (snakemake.threads, out, snakemake.params.prefix, rounds, out,
                                                   snakemake.params.prefix, rounds, out, snakemake.params.prefix, rounds,
                                                   out, snakemake.params.prefix, rounds), shell=True).wait()

    with open(os.path.join(out, "combined.%s.%d.pol.fa" % (snakemake.params.prefix, rounds)), "w") as o:
        with open(os.path.join(out, "filtered.%s.%d.pol.fa" % (snakemake.params.prefix, rounds))) as f:
            gotten_set = set()
            for line in f:
                if line.startswith('>'):
                    gotten_set.add(line.split()[0][1:])
                o.write(line)
        with open(reference) as f:
            for line in f:
                if line.startswith('>'):
                    name = line.split()[0][1:]
                    if name in gotten_set:
                        get_line = False
                    else:
                        get_line = True
                if get_line:
                    o.write(line)
    reference = os.path.join(out, "combined.%s.%d.pol.fa" % (snakemake.params.prefix, rounds))


if os.path.exists("data/short_reads.racon.1.fastq.gz"):
    os.remove("data/short_reads.racon.1.fastq.gz")
    # os.remove("data/short_reads.racon.2.fastq.gz")
shutil.copy2(reference, snakemake.output.fasta)
